import logging
import os.path
import shutil
import socket

import salt.config
import salt.modules.network as network
import salt.utils.path
from salt._compat import ipaddress
from salt.exceptions import CommandExecutionError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, mock_open, patch
from tests.support.unit import TestCase

log = logging.getLogger(__name__)


class NetworkTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.network
    """

    def setup_loader_modules(self):
        opts = salt.config.DEFAULT_MINION_OPTS.copy()
        utils = salt.loader.utils(
            opts, whitelist=["network", "path", "platform", "stringutils"]
        )
        return {
            network: {"__utils__": utils},
        }

    @patch("salt.utils.platform.is_windows")
    def test___virtual__is_windows_true(self, mock_is_windows):
        mock_is_windows.return_value = True
        result = network.__virtual__()
        expected = (
            False,
            "The network execution module cannot be loaded on Windows: use win_network"
            " instead.",
        )
        self.assertEqual(result, expected)

    @patch("salt.utils.platform.is_windows")
    def test___virtual__is_windows_false(self, mock_is_windows):
        mock_is_windows.return_value = False
        result = network.__virtual__()
        self.assertEqual(result, True)

    def test_wol_bad_mac(self):
        """
        tests network.wol with bad mac
        """
        bad_mac = "31337"
        self.assertRaises(ValueError, network.wol, bad_mac)

    def test_wol_success(self):
        """
        tests network.wol success
        """
        mac = "080027136977"
        bcast = "255.255.255.255 7"

        class MockSocket:
            def __init__(self, *args, **kwargs):
                pass

            def __call__(self, *args, **kwargs):
                pass

            def setsockopt(self, *args, **kwargs):
                pass

            def sendto(self, *args, **kwargs):
                pass

        with patch("socket.socket", MockSocket):
            self.assertTrue(network.wol(mac, bcast))

    def test_ping(self):
        """
        Test for Performs a ping to a host
        """
        with patch.dict(
            network.__utils__, {"network.sanitize_host": MagicMock(return_value="A")}
        ):
            mock_all = MagicMock(side_effect=[{"retcode": 1}, {"retcode": 0}])
            with patch.dict(network.__salt__, {"cmd.run_all": mock_all}):
                self.assertFalse(network.ping("host", return_boolean=True))

                self.assertTrue(network.ping("host", return_boolean=True))

            with patch.dict(network.__salt__, {"cmd.run": MagicMock(return_value="A")}):
                self.assertEqual(network.ping("host"), "A")

    def test_netstat(self):
        """
        Test for return information on open ports and states
        """
        with patch.dict(network.__grains__, {"kernel": "Linux"}):
            with patch.object(network, "_netstat_linux", return_value="A"):
                with patch.object(network, "_ss_linux", return_value="A"):
                    self.assertEqual(network.netstat(), "A")

        with patch.dict(network.__grains__, {"kernel": "OpenBSD"}):
            with patch.object(network, "_netstat_bsd", return_value="A"):
                self.assertEqual(network.netstat(), "A")

        with patch.dict(network.__grains__, {"kernel": "A"}):
            self.assertRaises(CommandExecutionError, network.netstat)

    def test_active_tcp(self):
        """
        Test for return a dict containing information on all
         of the running TCP connections
        """
        with patch.dict(
            network.__utils__, {"network.active_tcp": MagicMock(return_value="A")}
        ):
            with patch.dict(network.__grains__, {"kernel": "Linux"}):
                self.assertEqual(network.active_tcp(), "A")

    def test_traceroute(self):
        """
        Test for Performs a traceroute to a 3rd party host
        """

        def patched_which(binary):
            binary_path = shutil.which(binary)
            if binary_path:
                # The path exists, just return it
                return binary_path
            if binary == "traceroute":
                # The path doesn't exist but we mock it on the test.
                # Return the binary name
                return binary
            # The binary does not exist
            return binary_path

        with patch("salt.utils.path.which", patched_which):
            with patch.dict(network.__salt__, {"cmd.run": MagicMock(return_value="")}):
                self.assertListEqual(network.traceroute("gentoo.org"), [])

            with patch.dict(
                network.__utils__,
                {"network.sanitize_host": MagicMock(return_value="gentoo.org")},
            ):
                with patch.dict(
                    network.__salt__, {"cmd.run": MagicMock(return_value="")}
                ):
                    self.assertListEqual(network.traceroute("gentoo.org"), [])

    def test_dig(self):
        """
        Test for Performs a DNS lookup with dig
        """
        with patch("salt.utils.path.which", MagicMock(return_value="dig")), patch.dict(
            network.__utils__, {"network.sanitize_host": MagicMock(return_value="A")}
        ), patch.dict(network.__salt__, {"cmd.run": MagicMock(return_value="A")}):
            self.assertEqual(network.dig("host"), "A")

    def test_arp(self):
        """
        Test for return the arp table from the minion
        """
        with patch.dict(
            network.__salt__, {"cmd.run": MagicMock(return_value="A,B,C,D\nE,F,G,H\n")}
        ), patch("salt.utils.path.which", MagicMock(return_value="")):
            self.assertDictEqual(network.arp(), {})

    def test_interfaces(self):
        """
        Test for return a dictionary of information about
         all the interfaces on the minion
        """
        with patch.dict(
            network.__utils__, {"network.interfaces": MagicMock(return_value={})}
        ):
            self.assertDictEqual(network.interfaces(), {})

    def test_hw_addr(self):
        """
        Test for return the hardware address (a.k.a. MAC address)
         for a given interface
        """
        with patch.dict(
            network.__utils__, {"network.hw_addr": MagicMock(return_value={})}
        ):
            self.assertDictEqual(network.hw_addr("iface"), {})

    def test_interface(self):
        """
        Test for return the inet address for a given interface
        """
        with patch.dict(
            network.__utils__, {"network.interface": MagicMock(return_value={})}
        ):
            self.assertDictEqual(network.interface("iface"), {})

    def test_interface_ip(self):
        """
        Test for return the inet address for a given interface
        """
        with patch.dict(
            network.__utils__, {"network.interface_ip": MagicMock(return_value={})}
        ):
            self.assertDictEqual(network.interface_ip("iface"), {})

    def test_subnets(self):
        """
        Test for returns a list of subnets to which the host belongs
        """
        with patch.dict(
            network.__utils__, {"network.subnets": MagicMock(return_value={})}
        ):
            self.assertDictEqual(network.subnets(), {})

    def test_in_subnet(self):
        """
        Test for returns True if host is within specified
         subnet, otherwise False.
        """
        with patch.dict(
            network.__utils__, {"network.in_subnet": MagicMock(return_value={})}
        ):
            self.assertDictEqual(network.in_subnet("iface"), {})

    def test_ip_addrs(self):
        """
        Test for returns a list of IPv4 addresses assigned to the host.
        """
        with patch.dict(
            network.__utils__,
            {
                "network.ip_addrs": MagicMock(return_value=["0.0.0.0"]),
                "network.in_subnet": MagicMock(return_value=True),
            },
        ):
            self.assertListEqual(
                network.ip_addrs("interface", "include_loopback", "cidr"), ["0.0.0.0"]
            )
            self.assertListEqual(
                network.ip_addrs("interface", "include_loopback"), ["0.0.0.0"]
            )

    def test_ip_addrs6(self):
        """
        Test for returns a list of IPv6 addresses assigned to the host.
        """
        with patch.dict(
            network.__utils__, {"network.ip_addrs6": MagicMock(return_value=["A"])}
        ):
            self.assertListEqual(network.ip_addrs6("int", "include"), ["A"])

    def test_get_hostname(self):
        """
        Test for Get hostname
        """
        with patch.object(socket, "gethostname", return_value="A"):
            self.assertEqual(network.get_hostname(), "A")

    def test_mod_hostname(self):
        """
        Test for Modify hostname
        """
        self.assertFalse(network.mod_hostname(None))
        file_d = "\n".join(["#", "A B C D,E,F G H"])

        with patch.dict(
            network.__utils__,
            {
                "path.which": MagicMock(return_value="hostname"),
                "files.fopen": mock_open(read_data=file_d),
            },
        ), patch.dict(
            network.__salt__, {"cmd.run": MagicMock(return_value=None)}
        ), patch.dict(
            network.__grains__, {"os_family": "A"}
        ):
            self.assertTrue(network.mod_hostname("hostname"))

    def test_mod_hostname_quoted(self):
        """
        Test for correctly quoted hostname on rh-style distro
        """

        fopen_mock = mock_open(
            read_data={
                "/etc/hosts": "\n".join(
                    ["127.0.0.1 localhost.localdomain", "127.0.0.2 undef"]
                ),
                "/etc/sysconfig/network": "\n".join(
                    ["NETWORKING=yes", 'HOSTNAME="undef"']
                ),
            }
        )

        with patch.dict(network.__grains__, {"os_family": "RedHat"}), patch.dict(
            network.__salt__, {"cmd.run": MagicMock(return_value=None)}
        ), patch("socket.getfqdn", MagicMock(return_value="undef")), patch.dict(
            network.__utils__,
            {
                "path.which": MagicMock(return_value="hostname"),
                "files.fopen": fopen_mock,
            },
        ):
            self.assertTrue(network.mod_hostname("hostname"))
            assert (
                fopen_mock.filehandles["/etc/sysconfig/network"][1].write_calls[1]
                == 'HOSTNAME="hostname"\n'
            )

    def test_mod_hostname_unquoted(self):
        """
        Test for correctly unquoted hostname on rh-style distro
        """

        fopen_mock = mock_open(
            read_data={
                "/etc/hosts": "\n".join(
                    ["127.0.0.1 localhost.localdomain", "127.0.0.2 undef"]
                ),
                "/etc/sysconfig/network": "\n".join(
                    ["NETWORKING=yes", "HOSTNAME=undef"]
                ),
            }
        )

        with patch.dict(network.__grains__, {"os_family": "RedHat"}), patch.dict(
            network.__salt__, {"cmd.run": MagicMock(return_value=None)}
        ), patch("socket.getfqdn", MagicMock(return_value="undef")), patch.dict(
            network.__utils__,
            {
                "path.which": MagicMock(return_value="hostname"),
                "files.fopen": fopen_mock,
            },
        ):
            self.assertTrue(network.mod_hostname("hostname"))
            assert (
                fopen_mock.filehandles["/etc/sysconfig/network"][1].write_calls[1]
                == "HOSTNAME=hostname\n"
            )

    def test_connect(self):
        """
        Test for Test connectivity to a host using a particular
        port from the minion.
        """
        with patch("socket.socket") as mock_socket:
            self.assertDictEqual(
                network.connect(False, "port"),
                {"comment": "Required argument, host, is missing.", "result": False},
            )

            self.assertDictEqual(
                network.connect("host", False),
                {"comment": "Required argument, port, is missing.", "result": False},
            )

            ret = "Unable to connect to host (0) on tcp port port"
            mock_socket.side_effect = Exception("foo")
            with patch.dict(
                network.__utils__,
                {"network.sanitize_host": MagicMock(return_value="A")},
            ):
                with patch.object(
                    socket,
                    "getaddrinfo",
                    return_value=[["ipv4", "A", 6, "B", "0.0.0.0"]],
                ):
                    self.assertDictEqual(
                        network.connect("host", "port"),
                        {"comment": ret, "result": False},
                    )

            ret = "Successfully connected to host (0) on tcp port port"
            mock_socket.side_effect = MagicMock()
            mock_socket.settimeout().return_value = None
            mock_socket.connect().return_value = None
            mock_socket.shutdown().return_value = None
            with patch.dict(
                network.__utils__,
                {"network.sanitize_host": MagicMock(return_value="A")},
            ):
                with patch.object(
                    socket,
                    "getaddrinfo",
                    return_value=[["ipv4", "A", 6, "B", "0.0.0.0"]],
                ):
                    self.assertDictEqual(
                        network.connect("host", "port"),
                        {"comment": ret, "result": True},
                    )

    def test_is_private(self):
        """
        Test for Check if the given IP address is a private address
        """
        with patch.object(ipaddress.IPv4Address, "is_private", return_value=True):
            self.assertTrue(network.is_private("0.0.0.0"))
        with patch.object(ipaddress.IPv6Address, "is_private", return_value=True):
            self.assertTrue(network.is_private("::1"))

    def test_is_loopback(self):
        """
        Test for Check if the given IP address is a loopback address
        """
        with patch.object(ipaddress.IPv4Address, "is_loopback", return_value=True):
            self.assertTrue(network.is_loopback("127.0.0.1"))
        with patch.object(ipaddress.IPv6Address, "is_loopback", return_value=True):
            self.assertTrue(network.is_loopback("::1"))

    def test_get_bufsize(self):
        """
        Test for return network buffer sizes as a dict
        """
        with patch.dict(network.__grains__, {"kernel": "Linux"}):
            with patch.object(os.path, "exists", return_value=True):
                with patch.object(
                    network, "_get_bufsize_linux", return_value={"size": 1}
                ):
                    self.assertDictEqual(network.get_bufsize("iface"), {"size": 1})

        with patch.dict(network.__grains__, {"kernel": "A"}):
            self.assertDictEqual(network.get_bufsize("iface"), {})

    def test_mod_bufsize(self):
        """
        Test for Modify network interface buffers (currently linux only)
        """
        with patch.dict(network.__grains__, {"kernel": "Linux"}):
            with patch.object(os.path, "exists", return_value=True):
                with patch.object(
                    network, "_mod_bufsize_linux", return_value={"size": 1}
                ):
                    self.assertDictEqual(network.mod_bufsize("iface"), {"size": 1})

        with patch.dict(network.__grains__, {"kernel": "A"}):
            self.assertFalse(network.mod_bufsize("iface"))

    def test_routes(self):
        """
        Test for return currently configured routes from routing table
        """
        self.assertRaises(CommandExecutionError, network.routes, "family")

        with patch.dict(network.__grains__, {"kernel": "A", "os": "B"}):
            self.assertRaises(CommandExecutionError, network.routes, "inet")

        with patch.dict(network.__grains__, {"kernel": "Linux"}):
            with patch.object(
                network,
                "_netstat_route_linux",
                side_effect=["A", [{"addr_family": "inet"}]],
            ):
                with patch.object(
                    network,
                    "_ip_route_linux",
                    side_effect=["A", [{"addr_family": "inet"}]],
                ):
                    self.assertEqual(network.routes(None), "A")

                    self.assertListEqual(
                        network.routes("inet"), [{"addr_family": "inet"}]
                    )

    def test_default_route(self):
        """
        Test for return default route(s) from routing table
        """
        self.assertRaises(CommandExecutionError, network.default_route, "family")

        with patch.object(
            network,
            "routes",
            side_effect=[[{"addr_family": "inet"}, {"destination": "A"}], []],
        ):
            with patch.dict(network.__grains__, {"kernel": "A", "os": "B"}):
                self.assertRaises(CommandExecutionError, network.default_route, "inet")

            with patch.dict(network.__grains__, {"kernel": "Linux"}):
                self.assertListEqual(network.default_route("inet"), [])

    def test_default_route_ipv6(self):
        """
        Test for return default route(s) from routing table for IPv6
        Additionally tests that multicast, anycast, etc. do not throw errors
        """
        mock_iproute_ipv4 = """default via 192.168.0.1 dev enx3c18a040229d proto dhcp metric 100
default via 192.168.0.1 dev wlp59s0 proto dhcp metric 600
3.15.90.221 via 10.16.119.224 dev gpd0
3.18.18.213 via 10.16.119.224 dev gpd0
10.0.0.0/8 via 10.16.119.224 dev gpd0
10.1.0.0/16 via 10.12.240.1 dev tun0
10.2.0.0/16 via 10.12.240.1 dev tun0
10.12.0.0/16 via 10.12.240.1 dev tun0
10.12.240.0/20 dev tun0 proto kernel scope link src 10.12.240.2
10.14.0.0/16 via 10.12.240.1 dev tun0
10.16.0.0/16 via 10.12.240.1 dev tun0
10.16.188.201 via 10.16.119.224 dev gpd0
10.16.188.202 via 10.16.119.224 dev gpd0
10.27.0.0/16 via 10.12.240.1 dev tun0
52.14.149.204 via 10.16.119.224 dev gpd0
52.14.159.171 via 10.16.119.224 dev gpd0
52.14.249.61 via 10.16.119.224 dev gpd0
52.15.65.251 via 10.16.119.224 dev gpd0
54.70.229.135 via 10.16.119.224 dev gpd0
54.71.37.253 via 10.12.240.1 dev tun0
54.189.240.227 via 10.16.119.224 dev gpd0
66.170.96.2 via 192.168.0.1 dev enx3c18a040229d
80.169.184.191 via 10.16.119.224 dev gpd0
107.154.251.105 via 10.16.119.224 dev gpd0
168.61.48.213 via 10.16.119.224 dev gpd0
169.254.0.0/16 dev enx3c18a040229d scope link metric 1000
172.17.0.0/16 dev docker0 proto kernel scope link src 172.17.0.1 linkdown
172.30.0.0/16 via 10.12.240.1 dev tun0
184.169.136.236 via 10.16.119.224 dev gpd0
191.237.22.167 via 10.16.119.224 dev gpd0
192.30.68.16 via 10.16.119.224 dev gpd0
192.30.71.16 via 10.16.119.224 dev gpd0
192.30.71.71 via 10.16.119.224 dev gpd0
192.168.0.0/24 dev enx3c18a040229d proto kernel scope link src 192.168.0.99 metric 100
192.168.0.0/24 dev wlp59s0 proto kernel scope link src 192.168.0.99 metric 600
192.240.157.233 via 10.16.119.224 dev gpd0
206.80.50.33 via 10.16.119.224 dev gpd0
209.34.94.97 via 10.16.119.224 dev gpd0
unreachable should ignore this
"""
        mock_iproute_ipv6 = """::1 dev lo proto kernel metric 256 pref medium
2060:123:4069::10 dev enp5s0 proto kernel metric 100 pref medium
2060:123:4069::68 dev wlp3s0 proto kernel metric 600 pref medium
2060:123:4069::15:0/112 dev virbr0 proto kernel metric 256 pref medium
2060:123:4069::/64 dev enp5s0 proto ra metric 100 pref medium
2060:123:4069::/64 dev wlp3s0 proto ra metric 600 pref medium
2602:ae13:dc4:1b00::/56 via 2602:ae14:9e1:6080::10:1 dev tun0 proto static metric 50 pref medium
2602:ae14:66:8300::/56 via 2602:ae14:9e1:6080::10:1 dev tun0 proto static metric 50 pref medium
2602:ae14:a0:4d00::/56 via 2602:ae14:9e1:6080::10:1 dev tun0 proto static metric 50 pref medium
2602:ae14:508:3900::/56 via 2602:ae14:9e1:6080::10:1 dev tun0 proto static metric 50 pref medium
2602:ae14:513:a200::/56 via 2602:ae14:9e1:6080::10:1 dev tun0 proto static metric 50 pref medium
2602:ae14:769:2b00::/56 via 2602:ae14:9e1:6080::10:1 dev tun0 proto static metric 50 pref medium
2602:ae14:924:9700::/56 via 2602:ae14:9e1:6080::10:1 dev tun0 proto static metric 50 pref medium
2602:ae14:9e1:6000::10:1 via fe80::222:15ff:fe3f:23fe dev enp5s0 proto static metric 100 pref medium
2602:ae14:9e1:6080::10:1 dev tun0 proto kernel metric 50 pref medium
2602:ae14:9e1:6080::10:1 dev tun0 proto kernel metric 256 pref medium
2602:ae14:9e1:6080::10:1001 dev tun0 proto kernel metric 50 pref medium
2602:ae14:9e1:6000::/56 via 2602:ae14:9e1:6080::10:1 dev tun0 proto static metric 50 pref medium
2602:ae14:cc1:fa00::/56 via 2602:ae14:9e1:6080::10:1 dev tun0 proto static metric 50 pref medium
2602:ae14:cd0:5b00::/56 via 2602:ae14:9e1:6080::10:1 dev tun0 proto static metric 50 pref medium
2602:ae14:d5f:b400::/56 via 2602:ae14:9e1:6080::10:1 dev tun0 proto static metric 50 pref medium
2a34:d014:1d3:5d00::/56 via 2602:ae14:9e1:6080::10:1 dev tun0 proto static metric 50 pref medium
2a34:d014:919:bb00::/56 via 2602:ae14:9e1:6080::10:1 dev tun0 proto static metric 50 pref medium
fd0d:3ed3:cb42:1::/64 dev enp5s0 proto ra metric 100 pref medium
fd0d:3ed3:cb42:1::/64 dev wlp3s0 proto ra metric 600 pref medium
fe80::222:15ff:fe3f:23fe dev enp5s0 proto static metric 100 pref medium
fe80::/64 dev enp5s0 proto kernel metric 100 pref medium
fe80::/64 dev virbr0 proto kernel metric 256 pref medium
fe80::/64 dev vnet2 proto kernel metric 256 pref medium
fe80::/64 dev docker0 proto kernel metric 256 linkdown pref medium
fe80::/64 dev vpn0 proto kernel metric 256 pref medium
fe80::/64 dev wlp3s0 proto kernel metric 600 pref medium
default via fe80::222:15ff:fe3f:23fe dev enp5s0 proto ra metric 100 pref medium
default via fe80::222:15ff:fe3f:23fe dev wlp3s0 proto ra metric 600 pref medium
local ::1 dev lo table local proto kernel metric 0 pref medium
anycast 2060:123:4069:: dev wlp3s0 table local proto kernel metric 0 pref medium
local 2060:123:4069::10 dev enp5s0 table local proto kernel metric 0 pref medium
local 2060:123:4069::68 dev wlp3s0 table local proto kernel metric 0 pref medium
anycast 2060:123:4069::15:0 dev virbr0 table local proto kernel metric 0 pref medium
local 2060:123:4069::15:1 dev virbr0 table local proto kernel metric 0 pref medium
local 2060:123:4069:0:f4d:7d09:358c:ce5 dev wlp3s0 table local proto kernel metric 0 pref medium
local 2060:123:4069:0:a089:c284:32a8:9536 dev enp5s0 table local proto kernel metric 0 pref medium
anycast 2602:ae14:9e1:6080::10:0 dev tun0 table local proto kernel metric 0 pref medium
local 2602:ae14:9e1:6080::10:1001 dev tun0 table local proto kernel metric 0 pref medium
anycast fd0d:3ed3:cb42:1:: dev wlp3s0 table local proto kernel metric 0 pref medium
local fd0d:3ed3:cb42:1:cffd:9b03:c50:6d2a dev wlp3s0 table local proto kernel metric 0 pref medium
local fd0d:3ed3:cb42:1:f00b:50ef:2143:36cf dev enp5s0 table local proto kernel metric 0 pref medium
anycast fe80:: dev virbr0 table local proto kernel metric 0 pref medium
anycast fe80:: dev vnet2 table local proto kernel metric 0 pref medium
anycast fe80:: dev docker0 table local proto kernel metric 0 pref medium
anycast fe80:: dev wlp3s0 table local proto kernel metric 0 pref medium
anycast fe80:: dev vpn0 table local proto kernel metric 0 pref medium
local fe80::42:bfff:fec9:f590 dev docker0 table local proto kernel metric 0 pref medium
local fe80::18b1:cf8e:49cc:a783 dev wlp3s0 table local proto kernel metric 0 pref medium
local fe80::5054:ff:fe55:9457 dev virbr0 table local proto kernel metric 0 pref medium
local fe80::d251:c2a7:f5c8:2778 dev enp5s0 table local proto kernel metric 0 pref medium
local fe80::df35:e22c:f7db:a892 dev vpn0 table local proto kernel metric 0 pref medium
local fe80::fc54:ff:fee6:9fef dev vnet2 table local proto kernel metric 0 pref medium
multicast ff00::/8 dev enp5s0 table local proto kernel metric 256 pref medium
multicast ff00::/8 dev virbr0 table local proto kernel metric 256 pref medium
multicast ff00::/8 dev vnet2 table local proto kernel metric 256 pref medium
multicast ff00::/8 dev docker0 table local proto kernel metric 256 linkdown pref medium
multicast ff00::/8 dev wlp3s0 table local proto kernel metric 256 pref medium
multicast ff00::/8 dev vpn0 table local proto kernel metric 256 pref medium
multicast ff00::/8 dev tun0 table local proto kernel metric 256 pref medium
unicast should ignore this
broadcast cast should ignore this
throw should ignore this
unreachable should ignore this
prohibit should ignore this
blackhole should ignore this
nat should ignore this
"""

        self.assertRaises(CommandExecutionError, network.default_route, "family")

        with patch.object(
            network,
            "routes",
            side_effect=[[{"family": "inet6"}, {"destination": "A"}], []],
        ):
            with patch.dict(network.__grains__, {"kernel": "A", "os": "B"}):
                self.assertRaises(CommandExecutionError, network.default_route, "inet6")

        cmd_mock = MagicMock(side_effect=[mock_iproute_ipv4, mock_iproute_ipv6])
        with patch.dict(network.__grains__, {"kernel": "Linux"}):
            with patch.dict(
                network.__utils__, {"path.which": MagicMock(return_value=False)}
            ):
                with patch.dict(network.__salt__, {"cmd.run": cmd_mock}):
                    self.assertListEqual(
                        network.default_route("inet6"),
                        [
                            {
                                "addr_family": "inet6",
                                "destination": "::/0",
                                "gateway": "fe80::222:15ff:fe3f:23fe",
                                "netmask": "",
                                "flags": "UG",
                                "interface": "enp5s0",
                            },
                            {
                                "addr_family": "inet6",
                                "destination": "::/0",
                                "gateway": "fe80::222:15ff:fe3f:23fe",
                                "netmask": "",
                                "flags": "UG",
                                "interface": "wlp3s0",
                            },
                        ],
                    )

    def test_get_route(self):
        """
        Test for return output from get_route
        """
        mock_iproute = MagicMock(
            return_value="8.8.8.8 via 10.10.10.1 dev eth0 src 10.10.10.10 uid 0\ncache"
        )
        with patch.dict(network.__grains__, {"kernel": "Linux"}):
            with patch.dict(network.__salt__, {"cmd.run": mock_iproute}):
                expected = {
                    "interface": "eth0",
                    "source": "10.10.10.10",
                    "destination": "8.8.8.8",
                    "gateway": "10.10.10.1",
                }
                ret = network.get_route("8.8.8.8")
                self.assertEqual(ret, expected)

        mock_iproute = MagicMock(
            return_value=(
                "8.8.8.8 via 10.10.10.1 dev eth0.1 src 10.10.10.10 uid 0\ncache"
            )
        )
        with patch.dict(network.__grains__, {"kernel": "Linux"}):
            with patch.dict(network.__salt__, {"cmd.run": mock_iproute}):
                expected = {
                    "interface": "eth0.1",
                    "source": "10.10.10.10",
                    "destination": "8.8.8.8",
                    "gateway": "10.10.10.1",
                }
                ret = network.get_route("8.8.8.8")
                self.assertEqual(ret, expected)

        mock_iproute = MagicMock(
            return_value=(
                "8.8.8.8 via 10.10.10.1 dev eth0:1 src 10.10.10.10 uid 0\ncache"
            )
        )
        with patch.dict(network.__grains__, {"kernel": "Linux"}):
            with patch.dict(network.__salt__, {"cmd.run": mock_iproute}):
                expected = {
                    "interface": "eth0:1",
                    "source": "10.10.10.10",
                    "destination": "8.8.8.8",
                    "gateway": "10.10.10.1",
                }
                ret = network.get_route("8.8.8.8")
                self.assertEqual(ret, expected)

        mock_iproute = MagicMock(
            return_value=(
                "8.8.8.8 via 10.10.10.1 dev lan-br0 src 10.10.10.10 uid 0\ncache"
            )
        )
        with patch.dict(network.__grains__, {"kernel": "Linux"}):
            with patch.dict(network.__salt__, {"cmd.run": mock_iproute}):
                expected = {
                    "interface": "lan-br0",
                    "source": "10.10.10.10",
                    "destination": "8.8.8.8",
                    "gateway": "10.10.10.1",
                }
                ret = network.get_route("8.8.8.8")
                self.assertEqual(ret, expected)



"""
Tests for salt.modules.slack module
"""

import urllib.parse

import pytest

import salt.modules.slack_notify as slack_notify
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {slack_notify: {}}


def test_post_message():
    """
    Tests the post_message function sends data as expected
    """
    slack_query = MagicMock(return_value={"res": True})

    # bare minimum
    with patch("salt.utils.slack.query", slack_query):
        message_params = {
            "channel": "fake_channel",
            "from_name": "salt server",
            "message": "test message",
            "api_key": "xxx-xx-xxx",
        }
        assert slack_notify.post_message(**message_params)
        slack_query.assert_called_with(
            function="message",
            api_key="xxx-xx-xxx",
            method="POST",
            header_dict={"Content-Type": "application/x-www-form-urlencoded"},
            data=urllib.parse.urlencode(
                {
                    "channel": "#fake_channel",
                    "username": "salt server",
                    "text": "test message",
                    "attachments": [],
                    "blocks": [],
                }
            ),
            opts=slack_notify.__opts__,
        )

    # send `blocks` and `attachments` params
    with patch("salt.utils.slack.query", slack_query):
        message_params = {
            "channel": "fake_channel",
            "from_name": "salt server",
            "message": "test message",
            "api_key": "xxx-xx-xxx",
            "attachments": [{"text": "And heres an attachment!"}],
            "blocks": [
                {
                    "type": "header",
                    "text": {"type": "plain_text", "text": "New request"},
                }
            ],
        }
        assert slack_notify.post_message(**message_params)
        slack_query.assert_called_with(
            function="message",
            api_key="xxx-xx-xxx",
            method="POST",
            header_dict={"Content-Type": "application/x-www-form-urlencoded"},
            data=urllib.parse.urlencode(
                {
                    "channel": "#fake_channel",
                    "username": "salt server",
                    "text": "test message",
                    "attachments": [{"text": "And heres an attachment!"}],
                    "blocks": [
                        {
                            "type": "header",
                            "text": {"type": "plain_text", "text": "New request"},
                        }
                    ],
                }
            ),
            opts=slack_notify.__opts__,
        )



"""
Convenience functions for dealing with datetime classes
"""

import datetime

import salt.utils.stringutils
from salt.utils.decorators.jinja import jinja_filter

try:
    import timelib

    HAS_TIMELIB = True
except ImportError:
    HAS_TIMELIB = False


def date_cast(date):
    """
    Casts any object into a datetime.datetime object

    date
      any datetime, time string representation...
    """
    if date is None:
        return datetime.datetime.now()
    elif isinstance(date, datetime.datetime):
        return date

    # fuzzy date
    try:
        if isinstance(date, str):
            try:
                if HAS_TIMELIB:
                    # py3: yes, timelib.strtodatetime wants bytes, not str :/
                    return timelib.strtodatetime(salt.utils.stringutils.to_bytes(date))
            except ValueError:
                pass

            # not parsed yet, obviously a timestamp?
            if date.isdigit():
                date = int(date)
            else:
                date = float(date)

        return datetime.datetime.fromtimestamp(date)
    except Exception:  # pylint: disable=broad-except
        if HAS_TIMELIB:
            raise ValueError(f"Unable to parse {date}")

        raise RuntimeError(f"Unable to parse {date}. Consider installing timelib")


@jinja_filter("date_format")
@jinja_filter("strftime")
def strftime(date=None, format="%Y-%m-%d"):
    """
    Converts date into a time-based string

    date
      any datetime, time string representation...

    format
       :ref:`strftime<http://docs.python.org/2/library/datetime.html#datetime.datetime.strftime>` format

    >>> import datetime
    >>> src = datetime.datetime(2002, 12, 25, 12, 00, 00, 00)
    >>> strftime(src)
    '2002-12-25'
    >>> src = '2002/12/25'
    >>> strftime(src)
    '2002-12-25'
    >>> src = 1040814000
    >>> strftime(src)
    '2002-12-25'
    >>> src = '1040814000'
    >>> strftime(src)
    '2002-12-25'
    """
    return date_cast(date).strftime(format)


def total_seconds(td):
    """
    Takes a timedelta and returns the total number of seconds
    represented by the object. Wrapper for the total_seconds()
    method which does not exist in versions of Python < 2.7.
    """
    return (td.microseconds + (td.seconds + td.days * 24 * 3600) * 10**6) / 10**6



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.states.cloud as cloud
import salt.utils.cloud
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {cloud: {}}


def test_present():
    """
    Test to spin up a single instance on a cloud provider, using salt-cloud.
    """
    name = "mycloud"
    cloud_provider = "my_cloud_provider"

    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    mock = MagicMock(side_effect=[True, False])
    mock_bool = MagicMock(side_effect=[True, False, False])
    mock_dict = MagicMock(return_value={"cloud": "saltcloud"})
    with patch.dict(
        cloud.__salt__,
        {
            "cmd.retcode": mock,
            "cloud.has_instance": mock_bool,
            "cloud.create": mock_dict,
        },
    ):
        comt = "onlyif condition is false"
        ret.update({"comment": comt})
        assert cloud.present(name, cloud_provider, onlyif=False) == ret

        assert cloud.present(name, cloud_provider, onlyif="") == ret

        comt = "unless condition is true"
        ret.update({"comment": comt})
        assert cloud.present(name, cloud_provider, unless=True) == ret

        assert cloud.present(name, cloud_provider, unless="") == ret

        comt = f"Already present instance {name}"
        ret.update({"comment": comt})
        assert cloud.present(name, cloud_provider) == ret

        with patch.dict(cloud.__opts__, {"test": True}):
            comt = f"Instance {name} needs to be created"
            ret.update({"comment": comt, "result": None})
            assert cloud.present(name, cloud_provider) == ret

        with patch.dict(cloud.__opts__, {"test": False}):
            comt = (
                "Created instance mycloud using provider "
                "my_cloud_provider and the following options: {}"
            )
            ret.update(
                {"comment": comt, "result": True, "changes": {"cloud": "saltcloud"}}
            )
            assert cloud.present(name, cloud_provider) == ret


def test_absent():
    """
    Test to ensure that no instances with the specified names exist.
    """
    name = "mycloud"

    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    mock = MagicMock(side_effect=[True, False])
    mock_bool = MagicMock(side_effect=[False, True, True])
    mock_dict = MagicMock(return_value={"cloud": "saltcloud"})
    with patch.dict(
        cloud.__salt__,
        {
            "cmd.retcode": mock,
            "cloud.has_instance": mock_bool,
            "cloud.destroy": mock_dict,
        },
    ):
        comt = "onlyif condition is false"
        ret.update({"comment": comt})
        assert cloud.absent(name, onlyif=False) == ret

        assert cloud.absent(name, onlyif="") == ret

        comt = "unless condition is true"
        ret.update({"comment": comt})
        assert cloud.absent(name, unless=True) == ret

        assert cloud.absent(name, unless="") == ret

        comt = f"Already absent instance {name}"
        ret.update({"comment": comt})
        assert cloud.absent(name) == ret

        with patch.dict(cloud.__opts__, {"test": True}):
            comt = f"Instance {name} needs to be destroyed"
            ret.update({"comment": comt, "result": None})
            assert cloud.absent(name) == ret

        with patch.dict(cloud.__opts__, {"test": False}):
            comt = ("Destroyed instance {}").format(name)
            ret.update(
                {"comment": comt, "result": True, "changes": {"cloud": "saltcloud"}}
            )
            assert cloud.absent(name) == ret


def test_profile():
    """
    Test to create a single instance on a cloud provider,
    using a salt-cloud profile.
    """
    name = "mycloud"
    profile = "myprofile"

    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    mock = MagicMock(side_effect=[True, False])
    mock_dict = MagicMock(
        side_effect=[
            {"cloud": "saltcloud"},
            {"Not Actioned": True},
            {"Not Actioned": True},
            {"Not Found": True, "Not Actioned/Not Running": True},
        ]
    )
    mock_d = MagicMock(return_value={})
    with patch.dict(
        cloud.__salt__,
        {"cmd.retcode": mock, "cloud.profile": mock_d, "cloud.action": mock_dict},
    ):
        comt = "onlyif condition is false"
        ret.update({"comment": comt})
        assert cloud.profile(name, profile, onlyif=False) == ret

        assert cloud.profile(name, profile, onlyif="") == ret

        comt = "unless condition is true"
        ret.update({"comment": comt})
        assert cloud.profile(name, profile, unless=True) == ret

        assert cloud.profile(name, profile, unless="") == ret

        comt = f"Already present instance {name}"
        ret.update({"comment": comt})
        assert cloud.profile(name, profile) == ret

        with patch.dict(cloud.__opts__, {"test": True}):
            comt = f"Instance {name} needs to be created"
            ret.update({"comment": comt, "result": None})
            assert cloud.profile(name, profile) == ret

        with patch.dict(cloud.__opts__, {"test": False}):
            comt = f"Failed to create instance {name} using profile {profile}"
            ret.update({"comment": comt, "result": False})
            assert cloud.profile(name, profile) == ret

        with patch.dict(cloud.__opts__, {"test": False}):
            comt = f"Failed to create instance {name} using profile {profile}"
            ret.update({"comment": comt, "result": False})
            assert cloud.profile(name, profile) == ret


def test_volume_present():
    """
    Test to check that a block volume exists.
    """
    name = "mycloud"

    ret = {"name": name, "result": False, "changes": {}, "comment": ""}

    mock = MagicMock(return_value=name)
    mock_lst = MagicMock(side_effect=[[name], [], []])
    with patch.dict(
        cloud.__salt__, {"cloud.volume_list": mock_lst, "cloud.volume_create": mock}
    ):
        with patch.object(salt.utils.cloud, "check_name", MagicMock(return_value=True)):
            comt = "Invalid characters in name."
            ret.update({"comment": comt})
            assert cloud.volume_present(name) == ret

        comt = f"Volume exists: {name}"
        ret.update({"comment": comt, "result": True})
        assert cloud.volume_present(name) == ret

        with patch.dict(cloud.__opts__, {"test": True}):
            comt = f"Volume {name} will be created."
            ret.update({"comment": comt, "result": None})
            assert cloud.volume_present(name) == ret

        with patch.dict(cloud.__opts__, {"test": False}):
            comt = f"Volume {name} was created"
            ret.update(
                {
                    "comment": comt,
                    "result": True,
                    "changes": {"old": None, "new": name},
                }
            )
            assert cloud.volume_present(name) == ret


def test_volume_absent():
    """
    Test to check that a block volume exists.
    """
    name = "mycloud"

    ret = {"name": name, "result": False, "changes": {}, "comment": ""}

    mock = MagicMock(return_value=False)
    mock_lst = MagicMock(side_effect=[[], [name], [name]])
    with patch.dict(
        cloud.__salt__, {"cloud.volume_list": mock_lst, "cloud.volume_delete": mock}
    ):
        with patch.object(salt.utils.cloud, "check_name", MagicMock(return_value=True)):
            comt = "Invalid characters in name."
            ret.update({"comment": comt})
            assert cloud.volume_absent(name) == ret

        comt = "Volume is absent."
        ret.update({"comment": comt, "result": True})
        assert cloud.volume_absent(name) == ret

        with patch.dict(cloud.__opts__, {"test": True}):
            comt = f"Volume {name} will be deleted."
            ret.update({"comment": comt, "result": None})
            assert cloud.volume_absent(name) == ret

        with patch.dict(cloud.__opts__, {"test": False}):
            comt = f"Volume {name} failed to delete."
            ret.update({"comment": comt, "result": False})
            assert cloud.volume_absent(name) == ret


def test_volume_attached():
    """
    Test to check if a block volume is attached.
    """
    name = "mycloud"
    server_name = "mycloud_server"
    disk_name = "trogdor"

    ret = {"name": name, "result": False, "changes": {}, "comment": ""}

    mock = MagicMock(return_value=False)
    mock_dict = MagicMock(
        side_effect=[
            {name: {"name": disk_name, "attachments": True}},
            {},
            {name: {"name": disk_name, "attachments": False}},
            {name: {"name": disk_name, "attachments": False}},
            {name: {"name": disk_name, "attachments": False}},
        ]
    )
    with patch.dict(
        cloud.__salt__, {"cloud.volume_list": mock_dict, "cloud.action": mock}
    ):
        with patch.object(
            salt.utils.cloud,
            "check_name",
            MagicMock(side_effect=[True, False, True]),
        ):
            comt = "Invalid characters in name."
            ret.update({"comment": comt})
            assert cloud.volume_attached(name, server_name) == ret

            ret.update({"name": server_name})
            assert cloud.volume_attached(name, server_name) == ret

        comt = f"Volume {disk_name} is already attached: True"
        ret.update({"comment": comt, "result": True})
        assert cloud.volume_attached(name, server_name) == ret

        comt = f"Volume {name} does not exist"
        ret.update({"comment": comt, "result": False})
        assert cloud.volume_attached(name, server_name) == ret

        comt = f"Server {server_name} does not exist"
        ret.update({"comment": comt, "result": False})
        assert cloud.volume_attached(name, server_name) == ret

        mock = MagicMock(return_value=True)
        with patch.dict(
            cloud.__salt__, {"cloud.action": mock, "cloud.volume_attach": mock}
        ):
            with patch.dict(cloud.__opts__, {"test": True}):
                comt = f"Volume {name} will be will be attached."
                ret.update({"comment": comt, "result": None})
                assert cloud.volume_attached(name, server_name) == ret

            with patch.dict(cloud.__opts__, {"test": False}):
                comt = f"Volume {name} was created"
                ret.update(
                    {
                        "comment": comt,
                        "result": True,
                        "changes": {
                            "new": True,
                            "old": {"name": disk_name, "attachments": False},
                        },
                    }
                )
                assert cloud.volume_attached(name, server_name) == ret


def test_volume_detached():
    """
    Test to check if a block volume is detached.
    """
    name = "mycloud"
    server_name = "mycloud_server"
    disk_name = "trogdor"

    ret = {"name": name, "result": False, "changes": {}, "comment": ""}

    mock = MagicMock(return_value=False)
    mock_dict = MagicMock(
        side_effect=[
            {name: {"name": disk_name, "attachments": False}},
            {},
            {name: {"name": disk_name, "attachments": True}},
            {name: {"name": disk_name, "attachments": True}},
            {name: {"name": disk_name, "attachments": True}},
        ]
    )
    with patch.dict(
        cloud.__salt__, {"cloud.volume_list": mock_dict, "cloud.action": mock}
    ):
        with patch.object(
            salt.utils.cloud,
            "check_name",
            MagicMock(side_effect=[True, False, True]),
        ):
            comt = "Invalid characters in name."
            ret.update({"comment": comt})
            assert cloud.volume_detached(name, server_name) == ret

            ret.update({"name": server_name})
            assert cloud.volume_detached(name, server_name) == ret

        comt = f"Volume {disk_name} is not currently attached to anything."
        ret.update({"comment": comt, "result": True})
        assert cloud.volume_detached(name, server_name) == ret

        comt = f"Volume {name} does not exist"
        ret.update({"comment": comt})
        assert cloud.volume_detached(name, server_name) == ret

        comt = f"Server {server_name} does not exist"
        ret.update({"comment": comt})
        assert cloud.volume_detached(name, server_name) == ret

        mock = MagicMock(return_value=True)
        with patch.dict(
            cloud.__salt__, {"cloud.action": mock, "cloud.volume_detach": mock}
        ):
            with patch.dict(cloud.__opts__, {"test": True}):
                comt = f"Volume {name} will be will be detached."
                ret.update({"comment": comt, "result": None})
                assert cloud.volume_detached(name, server_name) == ret

            with patch.dict(cloud.__opts__, {"test": False}):
                comt = f"Volume {name} was created"
                ret.update(
                    {
                        "comment": comt,
                        "result": True,
                        "changes": {
                            "new": True,
                            "old": {"name": disk_name, "attachments": True},
                        },
                    }
                )
                assert cloud.volume_detached(name, server_name) == ret



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.modules.chef
import salt.states.chef as chef
from tests.support.mock import MagicMock, create_autospec, patch


@pytest.fixture
def configure_loader_modules():
    return {chef: {"__salt__": {}}}


@pytest.fixture(
    params=[
        ("chef.client", chef.client, create_autospec(salt.modules.chef.client)),
        ("chef.solo", chef.solo, create_autospec(salt.modules.chef.solo)),
    ]
)
def chef_state_and_mock_mod(request):
    mod_name, state, mock_mod = request.param
    with patch.dict(chef.__salt__, {mod_name: mock_mod}):
        yield state, mock_mod


@pytest.fixture
def test_mode():
    with patch.dict(chef.__opts__, {"test": True}):
        yield


@pytest.fixture
def not_test_mode():
    with patch.dict(chef.__opts__, {"test": False}):
        yield


@pytest.fixture(
    params=[
        "Chef Client finished, 1",
        "Chef Infra Client finished, 1",
        "Infra Phase complete, 1",
    ]
)
def changed_output(request):
    yield request.param


@pytest.fixture(
    params=[
        "Chef Client finished, 0",
        "Chef Infra Client finished, 0",
        "Infra Phase complete, 0",
    ]
)
def unchanged_output(request):
    yield request.param


@pytest.fixture
def successful_changed_output(changed_output, chef_state_and_mock_mod):
    _, mock_mod = chef_state_and_mock_mod
    mock_mod.return_value = {"retcode": 0, "stderr": "", "stdout": changed_output}
    yield


@pytest.fixture
def successful_unchanged_output(unchanged_output, chef_state_and_mock_mod):
    _, mock_mod = chef_state_and_mock_mod
    mock_mod.return_value = {"retcode": 0, "stderr": "", "stdout": unchanged_output}
    yield


@pytest.fixture
def unsuccessful_output(chef_state_and_mock_mod):
    _, mock_mod = chef_state_and_mock_mod
    mock_mod.return_value = {"retcode": 1, "stderr": "", "stdout": ""}
    yield


def test_client():
    """
    Test to run chef-client
    """
    name = "my-chef-run"

    ret = {"name": name, "result": False, "changes": {}, "comment": ""}

    mock = MagicMock(return_value={"retcode": 1, "stdout": "", "stderr": "error"})
    with patch.dict(chef.__salt__, {"chef.client": mock}):
        with patch.dict(chef.__opts__, {"test": True}):
            comt = "\nerror"
            ret.update({"comment": comt})
            assert chef.client(name) == ret


def test_solo():
    """
    Test to run chef-solo
    """
    name = "my-chef-run"

    ret = {"name": name, "result": False, "changes": {}, "comment": ""}

    mock = MagicMock(return_value={"retcode": 1, "stdout": "", "stderr": "error"})
    with patch.dict(chef.__salt__, {"chef.solo": mock}):
        with patch.dict(chef.__opts__, {"test": True}):
            comt = "\nerror"
            ret.update({"comment": comt})
            assert chef.solo(name) == ret


def test_when_testing_and_successful_changed_output_result_should_be_None(
    test_mode, successful_changed_output, chef_state_and_mock_mod
):
    state, _ = chef_state_and_mock_mod
    ret = state(name="fnord")
    assert ret["result"] is None


def test_when_testing_and_successful_unchanged_output_result_should_be_True(
    test_mode, successful_unchanged_output, chef_state_and_mock_mod
):
    state, _ = chef_state_and_mock_mod
    ret = state(name="fnord")
    assert ret["result"] is True


def test_when_not_testing_and_successful_changed_output_result_should_be_True(
    not_test_mode, successful_changed_output, chef_state_and_mock_mod
):
    state, _ = chef_state_and_mock_mod
    ret = state(name="fnord")
    assert ret["result"] is True


def test_when_not_testing_and_successful_unchanged_output_result_should_be_True(
    not_test_mode, successful_unchanged_output, chef_state_and_mock_mod
):
    state, _ = chef_state_and_mock_mod
    ret = state(name="fnord")
    assert ret["result"] is True


def test_when_not_testing_and_unsuccessful_output_result_should_be_False(
    not_test_mode, unsuccessful_output, chef_state_and_mock_mod
):
    state, _ = chef_state_and_mock_mod
    ret = state(name="fnord")
    assert ret["result"] is False



"""
unit tests for the slack engine
"""

import pytest

import salt.engines.slack as slack
from tests.support.mock import MagicMock, patch

pytestmark = [
    pytest.mark.skipif(
        slack.HAS_SLACKCLIENT is False, reason="The SlackClient is not installed"
    )
]


@pytest.fixture
def configure_loader_modules():
    return {slack: {}}


@pytest.fixture
def slack_client(minion_opts):
    token = "xoxb-xxxxxxxxxx-xxxxxxxxxxxxxxxxxxxxxxxx"

    with patch.dict(slack.__opts__, minion_opts):
        with patch("slackclient.SlackClient.rtm_connect", MagicMock(return_value=True)):
            slack_client = slack.SlackClient(token)
            yield slack_client


def test_control_message_target(slack_client):
    """
    Test slack engine: control_message_target
    """
    trigger_string = "!"

    loaded_groups = {
        "default": {
            "targets": {},
            "commands": {"cmd.run", "test.ping"},
            "default_target": {"tgt_type": "glob", "target": "*"},
            "users": {"gareth"},
            "aliases": {
                "whoami": {"cmd": "cmd.run whoami"},
                "list_pillar": {"cmd": "pillar.items"},
            },
        }
    }

    slack_user_name = "gareth"

    # Check for correct cmdline
    _expected = (True, {"tgt_type": "glob", "target": "*"}, ["cmd.run", "whoami"])
    text = "!cmd.run whoami"
    target_commandline = slack_client.control_message_target(
        slack_user_name, text, loaded_groups, trigger_string
    )

    assert target_commandline == _expected

    # Check aliases result in correct cmdline
    text = "!whoami"
    target_commandline = slack_client.control_message_target(
        slack_user_name, text, loaded_groups, trigger_string
    )

    assert target_commandline == _expected

    # Check pillar is overridden
    _expected = (
        True,
        {"tgt_type": "glob", "target": "*"},
        ["pillar.items", 'pillar={"hello": "world"}'],
    )
    text = r"""!list_pillar pillar='{"hello": "world"}'"""
    target_commandline = slack_client.control_message_target(
        slack_user_name, text, loaded_groups, trigger_string
    )

    assert target_commandline == _expected

    # Check target is overridden
    _expected = (
        True,
        {"tgt_type": "glob", "target": "localhost"},
        ["cmd.run", "whoami"],
    )
    text = "!cmd.run whoami target='localhost'"
    target_commandline = slack_client.control_message_target(
        slack_user_name, text, loaded_groups, trigger_string
    )

    assert target_commandline == _expected



import time

import pytest


@pytest.fixture
def start():
    return time.time()


def test_minion_start_event(
    start, event_listener, salt_master_1, salt_master_2, salt_minion_1
):
    start_events = event_listener.wait_for_events(
        [
            (salt_master_1.id, f"salt/minion/{salt_minion_1.id}/start"),
            (salt_master_2.id, f"salt/minion/{salt_minion_1.id}/start"),
        ],
        timeout=60,
        after_time=start,
    )
    assert not start_events.missed
    assert len(start_events.matches) == 2



"""
Glue execution module to link to the :mod:`fx2 proxymodule <salt.proxy.fx2>`.

Depends: :mod:`iDRAC Remote execution module (salt.modules.dracr) <salt.modules.dracr>`

For documentation on commands that you can direct to a Dell chassis via proxy,
look in the documentation for :mod:`salt.modules.dracr <salt.modules.dracr>`.

This execution module calls through to a function in the fx2 proxy module
called ``chconfig``.  That function looks up the function passed in the ``cmd``
parameter in :mod:`salt.modules.dracr <salt.modules.dracr>` and calls it.

.. versionadded:: 2015.8.2
"""

import logging

import salt.utils.platform

log = logging.getLogger(__name__)

__proxyenabled__ = ["fx2"]
__virtualname__ = "chassis"


def __virtual__():
    """
    Only work on proxy
    """
    if salt.utils.platform.is_proxy():
        return __virtualname__
    return (
        False,
        "The chassis execution module cannot be loaded: "
        "this only works in proxy minions.",
    )


def chassis_credentials():
    proxyprefix = __opts__["proxy"]["proxytype"]
    (username, password) = __proxy__[proxyprefix + ".find_credentials"]()
    return (username, password)


def cmd(cmd, *args, **kwargs):
    proxyprefix = __opts__["proxy"]["proxytype"]
    (username, password) = chassis_credentials()
    kwargs["admin_username"] = username
    kwargs["admin_password"] = password
    kwargs["host"] = __proxy__[proxyprefix + ".host"]()
    proxycmd = __opts__["proxy"]["proxytype"] + ".chconfig"
    return __proxy__[proxycmd](cmd, *args, **kwargs)



"""
An engine that reads messages from the salt event bus and pushes
them onto a logstash endpoint.

.. versionadded:: 2015.8.0

:configuration:

    Example configuration

    .. code-block:: yaml

        engines:
          - logstash:
              host: log.my_network.com
              port: 5959
              proto: tcp

:depends: logstash
"""

import logging

import salt.utils.event

try:
    import logstash
except ImportError:
    logstash = None

log = logging.getLogger(__name__)

__virtualname__ = "logstash"


def __virtual__():
    return (
        __virtualname__
        if logstash is not None
        else (False, "python-logstash not installed")
    )


def event_bus_context(opts):
    if opts.get("id").endswith("_master"):
        event_bus = salt.utils.event.get_master_event(
            opts, opts["sock_dir"], listen=True
        )
    else:
        event_bus = salt.utils.event.get_event(
            "minion",
            opts=opts,
            sock_dir=opts["sock_dir"],
            listen=True,
        )
    return event_bus


def start(host, port=5959, tag="salt/engine/logstash", proto="udp"):
    """
    Listen to salt events and forward them to logstash
    """

    if proto == "tcp":
        logstashHandler = logstash.TCPLogstashHandler
    elif proto == "udp":
        logstashHandler = logstash.UDPLogstashHandler

    logstash_logger = logging.getLogger("python-logstash-logger")
    logstash_logger.setLevel(logging.INFO)
    logstash_logger.addHandler(logstashHandler(host, port, version=1))

    with event_bus_context(__opts__) as event_bus:
        log.debug("Logstash engine started")
        while True:
            event = event_bus.get_event()
            if event:
                logstash_logger.info(tag, extra=event)



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.states.mongodb_database as mongodb_database
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {mongodb_database: {}}


def test_absent():
    """
    Test to ensure that the named database is absent.
    """
    name = "mydb"

    ret = {"name": name, "result": None, "comment": "", "changes": {}}

    mock = MagicMock(side_effect=[True, True, False])
    mock_t = MagicMock(return_value=True)
    with patch.dict(
        mongodb_database.__salt__,
        {"mongodb.db_exists": mock, "mongodb.db_remove": mock_t},
    ):
        with patch.dict(mongodb_database.__opts__, {"test": True}):
            comt = f"Database {name} is present and needs to be removed"
            ret.update({"comment": comt})
            assert mongodb_database.absent(name) == ret

        with patch.dict(mongodb_database.__opts__, {"test": False}):
            comt = f"Database {name} has been removed"
            ret.update({"comment": comt, "result": True, "changes": {"mydb": "Absent"}})
            assert mongodb_database.absent(name) == ret

            comt = f"Database {name} is not present"
            ret.update({"comment": comt, "changes": {}})
            assert mongodb_database.absent(name) == ret



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>
"""

import pytest

import salt.modules.bridge as bridge
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {bridge: {}}


def test_show():
    """
    Test for Returns bridges interfaces
    along with enslaved physical interfaces
    """
    mock = MagicMock(return_value=True)
    with patch.object(bridge, "_os_dispatch", mock):
        assert bridge.show("br")


def test_list_():
    """
    Test for Returns the machine's bridges list
    """
    mock = MagicMock(return_value=None)
    with patch.object(bridge, "_os_dispatch", mock):
        assert bridge.list_() is None

    mock = MagicMock(return_value=["A", "B"])
    with patch.object(bridge, "_os_dispatch", mock):
        assert bridge.list_() == ["A", "B"]


def test_interfaces():
    """
    Test for Returns interfaces attached to a bridge
    """
    assert bridge.interfaces() is None

    mock = MagicMock(return_value={"interfaces": "A"})
    with patch.object(bridge, "_os_dispatch", mock):
        assert bridge.interfaces("br") == "A"


def test_find_interfaces():
    """
    Test for Returns the bridge to which the interfaces are bond to
    """
    mock = MagicMock(return_value=None)
    with patch.object(bridge, "_os_dispatch", mock):
        assert bridge.find_interfaces() is None

    mock = MagicMock(return_value={"interfaces": "A"})
    with patch.object(bridge, "_os_dispatch", mock):
        assert bridge.find_interfaces() == {}


def test_add():
    """
    Test for Creates a bridge
    """
    mock = MagicMock(return_value="A")
    with patch.object(bridge, "_os_dispatch", mock):
        assert bridge.add() == "A"


def test_delete():
    """
    Test for Deletes a bridge
    """
    mock = MagicMock(return_value="A")
    with patch.object(bridge, "_os_dispatch", mock):
        assert bridge.delete() == "A"


def test_addif():
    """
    Test for Adds an interface to a bridge
    """
    mock = MagicMock(return_value="A")
    with patch.object(bridge, "_os_dispatch", mock):
        assert bridge.addif() == "A"


def test_delif():
    """
    Test for Removes an interface from a bridge
    """
    mock = MagicMock(return_value="A")
    with patch.object(bridge, "_os_dispatch", mock):
        assert bridge.delif() == "A"


def test_stp():
    """
    Test for Sets Spanning Tree Protocol state for a bridge
    """
    with patch.dict(bridge.__grains__, {"kernel": "Linux"}):
        mock = MagicMock(return_value="Linux")
        with patch.object(bridge, "_os_dispatch", mock):
            assert bridge.stp() == "Linux"

    with patch.dict(bridge.__grains__, {"kernel": "FreeBSD"}):
        mock = MagicMock(return_value="FreeBSD")
        with patch.object(bridge, "_os_dispatch", mock):
            assert bridge.stp() == "FreeBSD"

    with patch.dict(bridge.__grains__, {"kernel": None}):
        assert not bridge.stp()



import getpass
import logging
import os
import pathlib
import socket
import stat
import sys
import tempfile

import pytest

import salt.utils.files
import salt.utils.verify
from tests.support.mock import patch

if sys.platform.startswith("win"):
    import win32file
else:
    import resource

log = logging.getLogger(__name__)


@pytest.mark.skip_on_windows(reason="Not applicable for Windows.")
def test_verify_env_race_condition():
    def _stat(path):
        """
        Helper function for mock_stat, we want to raise errors for specific paths, but not until we get into the proper path.
        Until then, just return plain os.stat_result
        """
        if path in ("/tmp/salt-dir/.file3", "/tmp/salt-dir/.dir3"):
            raise AssertionError("The .file3 and .dir3 paths should never be called!")

        if path in ("/tmp/salt-dir/file1", "/tmp/salt-dir/dir1"):
            raise FileNotFoundError(
                "[Errno 2] No such file or directory: this exception should not be visible"
            )

        # we need to return at least different st_uid in order to trigger chown for these paths
        if path in ("/tmp/salt-dir/file4", "/tmp/salt-dir/dir4"):
            return os.stat_result([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11])

        return os.stat_result([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

    def _chown(path, uid, gid):
        if path in ("/tmp/salt-dir/file4", "/tmp/salt-dir/dir4"):
            raise FileNotFoundError(
                "[Errno 2] No such file or directory: this exception should not be visible"
            )

        return

    with patch("os.chown", side_effect=_chown) as mock_chown, patch(
        "os.stat", side_effect=_stat
    ) as mock_stat, patch(
        "salt.utils.verify._get_pwnam", return_value=(None, None, 0, 0)
    ), patch(
        "os.getuid", return_value=0
    ), patch(
        "os.listdir", return_value=["subdir"]
    ), patch(
        "os.path.isdir", return_value=True
    ), patch(
        "salt.utils.path.os_walk",
        return_value=[
            (
                "/tmp/salt-dir",
                ["dir1", "dir2", ".dir3", "dir4"],
                ["file1", "file2", ".file3", "file4"],
            )
        ],
    ):

        # verify this runs without issues, even though FNFE is raised
        salt.utils.verify.verify_env(["/tmp/salt-dir"], "root", skip_extra=True)

        # and verify it got actually called with the valid paths
        mock_stat.assert_any_call("/tmp/salt-dir/file1")
        mock_stat.assert_any_call("/tmp/salt-dir/dir1")

        mock_stat.assert_any_call("/tmp/salt-dir/file4")
        mock_stat.assert_any_call("/tmp/salt-dir/dir4")

        mock_chown.assert_any_call("/tmp/salt-dir/file4", 0, 0)
        mock_chown.assert_any_call("/tmp/salt-dir/dir4", 0, 0)


def test_valid_id_exception_handler():
    """
    Ensure we just return False if we pass in invalid or undefined paths.
    Refs #8259
    """
    opts = {"pki_dir": "/tmp/whatever"}
    assert not salt.utils.verify.valid_id(opts, None)


def test_valid_id_pathsep():
    """
    Path separators in id should make it invalid
    """
    opts = {"pki_dir": "/tmp/whatever"}
    # We have to test both path separators because os.path.normpath will
    # convert forward slashes to backslashes on Windows.
    for pathsep in ("/", "\\"):
        assert not salt.utils.verify.valid_id(opts, pathsep.join(("..", "foobar")))


def test_zmq_verify():
    assert salt.utils.verify.zmq_version()


def test_zmq_verify_insufficient():
    import zmq

    with patch.object(zmq, "__version__", "2.1.0"):
        assert not salt.utils.verify.zmq_version()


def test_user():
    assert salt.utils.verify.check_user(getpass.getuser())


def test_no_user():
    # Catch sys.stderr here since no logging is configured and
    # check_user WILL write to sys.stderr
    class FakeWriter:
        def __init__(self):
            self.output = ""
            self.errors = "strict"

        def write(self, data):
            self.output += data

        def flush(self):
            pass

    stderr = sys.stderr
    writer = FakeWriter()
    sys.stderr = writer
    try:
        # Now run the test
        if sys.platform.startswith("win"):
            assert salt.utils.verify.check_user("nouser")
        else:
            with pytest.raises(SystemExit):
                assert not salt.utils.verify.check_user("nouser")
    finally:
        # Restore sys.stderr
        sys.stderr = stderr
    if writer.output != 'CRITICAL: User not found: "nouser"\n':
        # If there's a different error catch, write it to sys.stderr
        sys.stderr.write(writer.output)


@pytest.mark.skip_on_windows(reason="No verify_env Windows")
def test_verify_env(tmp_path):
    root_dir = tmp_path / "root"
    var_dir = root_dir / "var" / "log" / "salt"
    key_dir = root_dir / "key_dir"
    salt.utils.verify.verify_env([var_dir], getpass.getuser(), root_dir=root_dir)
    assert var_dir.exists()
    dir_stat = os.stat(var_dir)
    assert dir_stat.st_uid == os.getuid()
    assert dir_stat.st_mode & stat.S_IRWXU == stat.S_IRWXU
    assert dir_stat.st_mode & stat.S_IRWXG == 40
    assert dir_stat.st_mode & stat.S_IRWXO == 5


@pytest.mark.requires_network(only_local_network=True)
def test_verify_socket():
    assert salt.utils.verify.verify_socket("", 18000, 18001)
    if socket.has_ipv6:
        # Only run if Python is built with IPv6 support; otherwise
        # this will just fail.
        try:
            assert salt.utils.verify.verify_socket("::", 18000, 18001)
        except OSError:
            # Python has IPv6 enabled, but the system cannot create
            # IPv6 sockets (otherwise the test would return a bool)
            # - skip the test
            #
            # FIXME - possibly emit a message that the system does
            # not support IPv6.
            pass


def test_max_open_files(caplog):
    with caplog.at_level(logging.DEBUG):
        recorded_logs = caplog.record_tuples
        logmsg_dbg = "This salt-master instance has accepted {0} minion keys."
        logmsg_chk = (
            "The number of accepted minion keys({}) should be lower "
            "than 1/4 of the max open files soft setting({}). According "
            "to the system's hard limit, there's still a margin of {} "
            "to raise the salt's max_open_files setting. Please consider "
            "raising this value."
        )
        logmsg_crash = (
            "The number of accepted minion keys({}) should be lower "
            "than 1/4 of the max open files soft setting({}). "
            "salt-master will crash pretty soon! According to the "
            "system's hard limit, there's still a margin of {} to "
            "raise the salt's max_open_files setting. Please consider "
            "raising this value."
        )
        if sys.platform.startswith("win"):
            logmsg_crash = (
                "The number of accepted minion keys({}) should be lower "
                "than 1/4 of the max open files soft setting({}). "
                "salt-master will crash pretty soon! Please consider "
                "raising this value."
            )

        if sys.platform.startswith("win"):
            # Check the Windows API for more detail on this
            # http://msdn.microsoft.com/en-us/library/xt874334(v=vs.71).aspx
            # and the python binding http://timgolden.me.uk/pywin32-docs/win32file.html
            mof_s = mof_h = win32file._getmaxstdio()
        else:
            mof_s, mof_h = resource.getrlimit(resource.RLIMIT_NOFILE)
        tempdir = tempfile.mkdtemp(prefix="fake-keys")
        keys_dir = pathlib.Path(tempdir, "minions")
        keys_dir.mkdir()

        mof_test = 256

        if sys.platform.startswith("win"):
            win32file._setmaxstdio(mof_test)
        else:
            resource.setrlimit(resource.RLIMIT_NOFILE, (mof_test, mof_h))

        try:
            prev = 0
            for newmax, level in (
                (24, None),
                (66, "INFO"),
                (127, "WARNING"),
                (196, "CRITICAL"),
            ):

                for n in range(prev, newmax):
                    kpath = pathlib.Path(keys_dir, str(n))
                    with salt.utils.files.fopen(kpath, "w") as fp_:
                        fp_.write(str(n))

                opts = {"max_open_files": newmax, "pki_dir": tempdir}

                salt.utils.verify.check_max_open_files(opts)

                if level is None:
                    # No log message is triggered, only the DEBUG one which
                    # tells us how many minion keys were accepted.
                    assert [logmsg_dbg.format(newmax)] == caplog.messages
                else:
                    assert logmsg_dbg.format(newmax) in caplog.messages
                    assert (
                        logmsg_chk.format(
                            newmax,
                            mof_test,
                            (
                                mof_test - newmax
                                if sys.platform.startswith("win")
                                else mof_h - newmax
                            ),
                        )
                        in caplog.messages
                    )
                prev = newmax

            newmax = mof_test
            for n in range(prev, newmax):
                kpath = pathlib.Path(keys_dir, str(n))
                with salt.utils.files.fopen(kpath, "w") as fp_:
                    fp_.write(str(n))

            opts = {"max_open_files": newmax, "pki_dir": tempdir}

            salt.utils.verify.check_max_open_files(opts)
            assert logmsg_dbg.format(newmax) in caplog.messages
            assert (
                logmsg_crash.format(
                    newmax,
                    mof_test,
                    (
                        mof_test - newmax
                        if sys.platform.startswith("win")
                        else mof_h - newmax
                    ),
                )
                in caplog.messages
            )
        except OSError as err:
            if err.errno == 24:
                # Too many open files
                pytest.skip("We've hit the max open files setting")
            raise
        finally:
            if sys.platform.startswith("win"):
                win32file._setmaxstdio(mof_h)
            else:
                resource.setrlimit(resource.RLIMIT_NOFILE, (mof_s, mof_h))



"""
Test the hosts module
"""

import logging
import os
import shutil

import pytest

import salt.utils.files
import salt.utils.stringutils
from tests.support.case import ModuleCase
from tests.support.runtests import RUNTIME_VARS

log = logging.getLogger(__name__)


@pytest.mark.windows_whitelisted
class HostsModuleTest(ModuleCase):
    """
    Test the hosts module
    """

    maxDiff = None

    @classmethod
    def setUpClass(cls):
        cls.hosts_file = os.path.join(RUNTIME_VARS.TMP, "hosts")

    def __clear_hosts(self):
        """
        Delete the tmp hosts file
        """
        if os.path.isfile(self.hosts_file):
            os.remove(self.hosts_file)

    def setUp(self):
        shutil.copyfile(os.path.join(RUNTIME_VARS.FILES, "hosts"), self.hosts_file)
        self.addCleanup(self.__clear_hosts)

    @pytest.mark.slow_test
    def test_list_hosts(self):
        """
        hosts.list_hosts
        """
        hosts = self.run_function("hosts.list_hosts")
        self.assertEqual(len(hosts), 10)
        self.assertEqual(hosts["::1"], {"aliases": ["ip6-localhost", "ip6-loopback"]})
        self.assertEqual(hosts["127.0.0.1"], {"aliases": ["localhost", "myname"]})

    @pytest.mark.slow_test
    def test_list_hosts_nofile(self):
        """
        hosts.list_hosts
        without a hosts file
        """
        if os.path.isfile(self.hosts_file):
            os.remove(self.hosts_file)
        hosts = self.run_function("hosts.list_hosts")
        self.assertEqual(hosts, {})

    @pytest.mark.slow_test
    def test_get_ip(self):
        """
        hosts.get_ip
        """
        self.assertEqual(self.run_function("hosts.get_ip", ["myname"]), "127.0.0.1")
        self.assertEqual(self.run_function("hosts.get_ip", ["othername"]), "")
        self.__clear_hosts()
        self.assertEqual(self.run_function("hosts.get_ip", ["othername"]), "")

    @pytest.mark.slow_test
    def test_get_alias(self):
        """
        hosts.get_alias
        """
        self.assertEqual(
            self.run_function("hosts.get_alias", ["127.0.0.1"]), ["localhost", "myname"]
        )
        self.assertEqual(self.run_function("hosts.get_alias", ["127.0.0.2"]), [])
        self.__clear_hosts()
        self.assertEqual(self.run_function("hosts.get_alias", ["127.0.0.1"]), [])

    @pytest.mark.slow_test
    def test_has_pair(self):
        """
        hosts.has_pair
        """
        self.assertTrue(self.run_function("hosts.has_pair", ["127.0.0.1", "myname"]))
        self.assertFalse(
            self.run_function("hosts.has_pair", ["127.0.0.1", "othername"])
        )

    @pytest.mark.slow_test
    def test_set_host(self):
        """
        hosts.set_hosts
        """
        self.assertTrue(self.run_function("hosts.set_host", ["192.168.1.123", "newip"]))
        self.assertTrue(self.run_function("hosts.has_pair", ["192.168.1.123", "newip"]))
        self.assertTrue(self.run_function("hosts.set_host", ["127.0.0.1", "localhost"]))
        self.assertEqual(len(self.run_function("hosts.list_hosts")), 11)
        self.assertFalse(
            self.run_function("hosts.has_pair", ["127.0.0.1", "myname"]),
            "should remove second entry",
        )

    @pytest.mark.slow_test
    def test_add_host(self):
        """
        hosts.add_host
        """
        self.assertTrue(self.run_function("hosts.add_host", ["192.168.1.123", "newip"]))
        self.assertTrue(self.run_function("hosts.has_pair", ["192.168.1.123", "newip"]))
        self.assertEqual(len(self.run_function("hosts.list_hosts")), 11)
        self.assertTrue(
            self.run_function("hosts.add_host", ["127.0.0.1", "othernameip"])
        )
        self.assertEqual(len(self.run_function("hosts.list_hosts")), 11)

    @pytest.mark.slow_test
    def test_rm_host(self):
        self.assertTrue(self.run_function("hosts.has_pair", ["127.0.0.1", "myname"]))
        self.assertTrue(self.run_function("hosts.rm_host", ["127.0.0.1", "myname"]))
        self.assertFalse(self.run_function("hosts.has_pair", ["127.0.0.1", "myname"]))
        self.assertTrue(self.run_function("hosts.rm_host", ["127.0.0.1", "unknown"]))

    @pytest.mark.slow_test
    def test_add_host_formatting(self):
        """
        Ensure that hosts.add_host isn't adding duplicates and that
        it's formatting the output correctly
        """
        # instead of using the 'clean' hosts file we're going to
        # use an empty one so we can prove the syntax of the entries
        # being added by the hosts module
        self.__clear_hosts()
        with salt.utils.files.fopen(self.hosts_file, "w"):
            pass

        self.assertTrue(
            self.run_function("hosts.add_host", ["192.168.1.3", "host3.fqdn.com"])
        )
        self.assertTrue(
            self.run_function("hosts.add_host", ["192.168.1.1", "host1.fqdn.com"])
        )
        self.assertTrue(self.run_function("hosts.add_host", ["192.168.1.1", "host1"]))
        self.assertTrue(
            self.run_function("hosts.add_host", ["192.168.1.2", "host2.fqdn.com"])
        )
        self.assertTrue(self.run_function("hosts.add_host", ["192.168.1.2", "host2"]))
        self.assertTrue(
            self.run_function("hosts.add_host", ["192.168.1.2", "oldhost2"])
        )
        self.assertTrue(
            self.run_function("hosts.add_host", ["192.168.1.2", "host2-reorder"])
        )
        self.assertTrue(
            self.run_function("hosts.add_host", ["192.168.1.1", "host1-reorder"])
        )

        # now read the lines and ensure they're formatted correctly
        with salt.utils.files.fopen(self.hosts_file, "r") as fp_:
            lines = salt.utils.stringutils.to_unicode(fp_.read()).splitlines()
        self.assertEqual(
            lines,
            [
                "192.168.1.3\t\thost3.fqdn.com",
                "192.168.1.1\t\thost1.fqdn.com host1 host1-reorder",
                "192.168.1.2\t\thost2.fqdn.com host2 oldhost2 host2-reorder",
            ],
        )



"""
An engine that uses presence detection to keep track of which minions
have been recently connected and remove their keys if they have not been
connected for a certain period of time.

Requires that the :conf_master:`minion_data_cache` option be enabled.

.. versionadded:: 2017.7.0

:configuration:

    Example configuration:

    .. code-block:: yaml

        engines:
          - stalekey:
              interval: 3600
              expire: 86400

"""

import logging
import os
import time

import salt.config
import salt.key
import salt.utils.files
import salt.utils.minions
import salt.utils.msgpack
import salt.wheel

log = logging.getLogger(__name__)


def __virtual__():
    if not __opts__.get("minion_data_cache"):
        return (False, "stalekey engine requires minion_data_cache to be enabled")
    return True


def _get_keys():
    """
    Get the keys
    """
    with salt.key.get_key(__opts__) as keys:
        minions = keys.all_keys()
        return minions["minions"]


def _delete_keys(stale_keys, minions):
    """
    Delete the keys
    """
    wheel = salt.wheel.WheelClient(__opts__)
    for k in stale_keys:
        log.info("Removing stale key for %s", k)
        wheel.cmd("key.delete", [salt.utils.stringutils.to_unicode(k)])
        del minions[k]
    return minions


def _read_presence(presence_file):
    """
    Read minion data from presence file
    """
    error = False
    minions = {}
    if os.path.exists(presence_file):
        try:
            with salt.utils.files.fopen(presence_file, "rb") as f:
                _minions = salt.utils.msgpack.load(f)

                # ensure all keys are unicode, not bytes.
                for minion in _minions:
                    _minion = salt.utils.stringutils.to_unicode(minion)
                    minions[_minion] = _minions[minion]

        except OSError as e:
            error = True
            log.error("Could not open presence file %s: %s", presence_file, e)

    return error, minions


def _write_presence(presence_file, minions):
    """
    Write minion data to presence file
    """
    error = False
    try:
        with salt.utils.files.fopen(presence_file, "wb") as f:
            salt.utils.msgpack.dump(minions, f)
    except OSError as e:
        error = True
        log.error("Could not write to presence file %s: %s", presence_file, e)
    return error


def start(interval=3600, expire=604800):
    """
    Start the engine
    """
    ck = salt.utils.minions.CkMinions(__opts__)
    presence_file = "{}/presence.p".format(__opts__["cachedir"])
    wheel = salt.wheel.WheelClient(__opts__)

    while True:
        log.debug("Checking for present minions")
        minions = {}
        error, minions = _read_presence(presence_file)
        if error:
            time.sleep(interval)
            continue

        minion_keys = _get_keys()
        now = time.time()
        present = ck.connected_ids()

        # For our existing keys, check which are present
        for m in minion_keys:
            # If we have a key that's not in the presence file,
            # it may be a new minion # It could also mean this
            # is the first time this engine is running and no
            # presence file was found
            if m not in minions:
                minions[m] = now
            elif m in present:
                minions[m] = now

        log.debug("Finished checking for present minions")
        # Delete old keys
        stale_keys = []
        for m, seen in minions.items():
            if now - expire > seen:
                stale_keys.append(m)

        if stale_keys:
            minions = _delete_keys(stale_keys, minions)

        error = _write_presence(presence_file, minions)

        time.sleep(interval)



import logging
import pathlib
import shutil
import subprocess
import textwrap

import pytest
from pytestshellutils.utils.processes import ProcessResult

log = logging.getLogger(__name__)


# The following fixtures are copied from pytests/functional/pillar/test_gpg.py


@pytest.fixture(scope="module")
def test_key():
    """
    Private key for setting up GPG pillar environment.
    """
    return textwrap.dedent(
        """\
        -----BEGIN PGP PRIVATE KEY BLOCK-----

        lQOYBFiKrcYBCADAj92+fz20uKxxH0ffMwcryGG9IogkiUi2QrNYilB4hwrY5Qt7
        Sbywlk/mSDMcABxMxS0vegqc5pgglvAnsi9w7j//9nfjiirsyiTYOOD1akTFQr7b
        qT6zuGFA4oYmYHvfBOena485qvlyitYLKYT9h27TDiiH6Jgt4xSRbjeyhTf3/fKD
        JzHA9ii5oeVi1pH/8/4USgXanBdKwO0JKQtci+PF0qe/nkzRswqTIkdgx1oyNUqL
        tYJ0XPOy+UyOC4J4QDIt9PQbAmiur8By4g2lLYWlGOCjs7Fcj3n5meWKzf1pmXoY
        lAnSab8kUZSSkoWQoTO7RbjFypULKCZui45/ABEBAAEAB/wM1wsAMtfYfx/wgxd1
        yJ9HyhrKU80kMotIq/Xth3uKLecJQ2yakfYlCEDXqCTQTymT7OnwaoDeqXmnYqks
        3HLRYvGdjb+8ym/GTkxapqBJfQaM6MB1QTnPHhJOE0zCrlhULK2NulxYihAMFTnk
        kKYviaJYLG+DcH0FQkkS0XihTKcqnsoJiS6iNd5SME3pa0qijR0D5f78fkvNzzEE
        9vgAX1TgQ5PDJGN6nYlW2bWxTcg+FR2cUAQPTiP9wXCH6VyJoQay7KHVr3r/7SsU
        89otfcx5HVDYPrez6xnP6wN0P/mKxCDbkERLDjZjWOmNXg2zn+/t3u02e+ybfAIp
        kTTxBADY/FmPgLpJ2bpcPH141twpHwhKIbENlTB9745Qknr6aLA0QVCkz49/3joO
        Sj+SZ7Jhl6cfbynrfHwX3b1bOFTzBUH2Tsi0HX40PezEFH0apf55FLZuMOBt/lc1
        ET6evpIHF0dcM+BvZa7E7MyTyEq8S7Cc9RoJyfeGbS7MG5FfuwQA4y9QOb/OQglq
        ZffkVItwY52RKWb/b2WQmt+IcVax/j7DmBva765SIfPDvOCMrYhJBI/uYHQ0Zia7
        SnC9+ez55wdYqgHkYojc21CIOnUvsPSj+rOpryoXzmcTuvKeVIyIA0h/mQyWjimR
        ENrikC4+O8GBMY6V4uvS4EFhLfHE9g0D/20lNOKkpAKPenr8iAPWcl0/pijJCGxF
        agnT7O2GQ9Lr5hSjW86agkevbGktu2ja5t/fHq0wpLQ4DVLMrR0/poaprTr307kW
        AlQV3z/C2cMHNysz4ulOgQrudQbhUEz2A8nQxRtIfWunkEugKLr1QiCkE1LJW8Np
        ZLxE6Qp0/KzdQva0HVNhbHQgR1BHIDxlcmlrQHNhbHRzdGFjay5jb20+iQFUBBMB
        CAA+FiEE+AxQ1ELHGEyFTZPYw5x3k9EbHGsFAliKrcYCGwMFCQPCZwAFCwkIBwIG
        FQgJCgsCBBYCAwECHgECF4AACgkQw5x3k9EbHGubUAf+PLdp1oTLVokockZgLyIQ
        wxOd3ofNOgNk4QoAkSMNSbtnYoQFKumRw/yGyPSIoHMsOC/ga98r8TAJEKfx3DLA
        rsD34oMAaYUT+XUd0KoSmlHqBrtDD1+eBASKYsCosHpCiKuQFfLKSxvpEr2YyL8L
        X3Q2TY5zFlGA9Eeq5g+rlb++yRZrruFN28EWtY/pyXFZgIB30ReDwPkM9hrioPZM
        0Qf3+dWZSK1rWViclB51oNy4un9stTiFZptAqz4NTNssU5A4AcNQPwBwnKIYoE58
        Y/Zyv8HzILGykT+qFebqRlRBI/13eHdzgJOL1iPRfjTk5Cvr+vcyIxAklXOP81ja
        B50DmARYiq3GAQgArnzu4SPCCQGNcCNxN4QlMP5TNvRsm5KrPbcO9j8HPfB+DRXs
        6B3mnuR6OJg7YuC0C2A/m2dSHJKkF0f2AwFRpxLjJ2iAFbrZAW/N0vZDx8zO+YAU
        HyLu0V04wdCE5DTLkgfWNR+0uMa8qZ4Kn56Gv7O+OFE7zgTHeZ7psWlxdafeW7u6
        zlC/3DWksNtuNb0vQDNMM4vgXbnORIfXdyh41zvEEnr/rKw8DuJAmo20mcv6Qi51
        PqqyM62ddQOEVfiMs9l4vmwZAjGFNFNInyPXnogL6UPCDmizb6hh8aX/MwG/XFIG
        KMJWbAVGpyBuqljKIt3qLu/s8ouPqkEN+f+nGwARAQABAAf+NA36d/kieGxZpTQ1
        oQHP1Jty+OiXhBwP8SPtF0J7ZxuZh07cs+zDsfBok/y6bsepfuFSaIq84OBQis+B
        kajxkp3cXZPb7l+lQLv5k++7Dd7Ien+ewSE7TQN6HLwYATrM5n5nBcc1M5C6lQGc
        mr0A5yz42TVG2bHsTpi9kBtsaVRSPUHSh8A8T6eOyCrT+/CAJVEEf7JyNyaqH1dy
        LuxI1VF3ySDEtFzuwN8EZQP9Yz/4AVyEQEA7WkNEwSQsBi2bWgWEdG+qjqnL+YKa
        vwe7/aJYPeL1zICnP/Osd/UcpDxR78MbozstbRljML0fTLj7UJ+XDazwv+Kl0193
        2ZK2QQQAwgXvS19MYNkHO7kbNVLt1VE2ll901iC9GFHBpFUam6gmoHXpCarB+ShH
        8x25aoUu4MxHmFxXd+Zq3d6q2yb57doWoPgvqcefpGmigaITnb1jhV2rt65V8deA
        SQazZNqBEBbZNIhfn6ObxHXXvaYaqq/UOEQ7uKyR9WMJT/rmqMEEAOY5h1R1t7AB
        JZ5VnhyAhdsNWw1gTcXB3o8gKz4vjdnPm0F4aVIPfB3BukETDc3sc2tKmCfUF7I7
        oOrh7iRez5F0RIC3KDzXF8qUuWBfPViww45JgftdKsecCIlEEYCoc+3goX0su2bP
        V1MDuHijMGTJCBABDgizNb0oynW5xcrbA/0QnKfpTwi7G3oRcJWv2YebVDRcU+SP
        dOYhq6SnmWPizEIljRG/X7FHJB+W7tzryO3sCDTAYwxFrfMwvJ2PwnAYI4349zYd
        lC28HowUkBYNhwBXc48xCfyhPZtD0aLx/OX1oLZ/vi8gd8TusgGupV/JjkFVO+Nd
        +shN/UEAldwqkkY2iQE8BBgBCAAmFiEE+AxQ1ELHGEyFTZPYw5x3k9EbHGsFAliK
        rcYCGwwFCQPCZwAACgkQw5x3k9EbHGu4wwf/dRFat91BRX1TJfwJl5otoAXpItYM
        6kdWWf1Eb1BicAvXhI078MSH4WXdKkJjJr1fFP8Ynil513H4Mzb0rotMAhb0jLSA
        lSRkMbhMvPxoS2kaYzioaBpp8yXpGiNo7dF+PJXSm/Uwp3AkcFjoVbBOqDWGgxMi
        DvDAstzLZ9dIcmr+OmcRQykKOKXlhEl3HnR5CyuPrA8hdVup4oeVwdkJhfJFKLLb
        3fR26wxJOmIOAt24eAUy721WfQ9txNAmhdy8mY842ODZESw6WatrQjRfuqosDgrk
        jc0cCHsEqJNZ2AB+1uEl3tcH0tyAFJa33F0znSonP17SS1Ff9sgHYBVLUg==
        =06Tz
        -----END PGP PRIVATE KEY BLOCK-----
        """
    )


@pytest.fixture(scope="module")
def gpg_pillar_yaml():
    """
    Yaml data for testing GPG pillar.
    """
    return textwrap.dedent(
        """
        #!yaml|gpg
        secrets:
          foo: |
            -----BEGIN PGP MESSAGE-----

            hQEMAw2B674HRhwSAQgAhTrN8NizwUv/VunVrqa4/X8t6EUulrnhKcSeb8sZS4th
            W1Qz3K2NjL4lkUHCQHKZVx/VoZY7zsddBIFvvoGGfj8+2wjkEDwFmFjGE4DEsS74
            ZLRFIFJC1iB/O0AiQ+oU745skQkU6OEKxqavmKMrKo3rvJ8ZCXDC470+i2/Hqrp7
            +KWGmaDOO422JaSKRm5D9bQZr9oX7KqnrPG9I1+UbJyQSJdsdtquPWmeIpamEVHb
            VMDNQRjSezZ1yKC4kCWm3YQbBF76qTHzG1VlLF5qOzuGI9VkyvlMaLfMibriqY73
            zBbPzf6Bkp2+Y9qyzuveYMmwS4sEOuZL/PetqisWe9JGAWD/O+slQ2KRu9hNww06
            KMDPJRdyj5bRuBVE4hHkkP23KrYr7SuhW2vpe7O/MvWEJ9uDNegpMLhTWruGngJh
            iFndxegN9w==
            =bAuo
            -----END PGP MESSAGE-----
        """
    )


@pytest.fixture(scope="module")
def gpg_homedir(salt_master, test_key):
    """
    Setup gpg environment
    """
    _gpg_homedir = pathlib.Path(salt_master.config_dir) / "gpgkeys"
    _gpg_homedir.mkdir(0o700)
    agent_started = False
    try:
        cmd_prefix = ["gpg", "--homedir", str(_gpg_homedir)]

        cmd = cmd_prefix + ["--list-keys"]
        proc = subprocess.run(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            check=True,
            text=True,
        )
        ret = ProcessResult(
            returncode=proc.returncode,
            stdout=proc.stdout,
            stderr=proc.stderr or "",
            cmdline=proc.args,
        )
        log.debug("Instantiating gpg keyring...\n%s", ret)

        cmd = cmd_prefix + ["--import", "--allow-secret-key-import"]
        proc = subprocess.run(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            check=True,
            text=True,
            input=test_key,
        )
        ret = ProcessResult(
            returncode=proc.returncode,
            stdout=proc.stdout,
            stderr=proc.stderr or "",
            cmdline=proc.args,
        )
        log.debug("Importing keypair...:\n%s", ret)

        agent_started = True

        yield _gpg_homedir
    finally:
        if agent_started:
            try:
                cmd = ["gpg-connect-agent", "--homedir", str(_gpg_homedir)]
                proc = subprocess.run(
                    cmd,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                    check=True,
                    text=True,
                    input="KILLAGENT",
                )
                ret = ProcessResult(
                    returncode=proc.returncode,
                    stdout=proc.stdout,
                    stderr=proc.stderr or "",
                    cmdline=proc.args,
                )
                log.debug("Killed gpg-agent...\n%s", ret)
            except (OSError, subprocess.CalledProcessError):
                log.debug("No need to kill: old gnupg doesn't start the agent.")
        shutil.rmtree(str(_gpg_homedir), ignore_errors=True)


@pytest.fixture(scope="module")
def pillar_setup(base_env_pillar_tree_root_dir, gpg_pillar_yaml, salt_minion):
    """
    Setup gpg pillar
    """
    saltutil_contents = f"""
    saltutil: {{{{ salt["saltutil.runner"]("mine.get", tgt="{salt_minion.id}", fun="test.ping") | json }}}}
    """
    top_file_contents = """
    base:
      '*':
        - gpg
        - saltutil
    """
    with pytest.helpers.temp_file(
        "top.sls", top_file_contents, base_env_pillar_tree_root_dir
    ), pytest.helpers.temp_file(
        "gpg.sls", gpg_pillar_yaml, base_env_pillar_tree_root_dir
    ), pytest.helpers.temp_file(
        "saltutil.sls", saltutil_contents, base_env_pillar_tree_root_dir
    ):
        yield


@pytest.mark.skip_if_binaries_missing("gpg")
@pytest.mark.usefixtures("pillar_setup", "gpg_homedir")
def test_gpg_pillar(salt_ssh_cli):
    """
    Ensure that GPG-encrypted pillars can be decrypted, i.e. the
    gpg_keydir should not be overridden. This is issue #60002,
    which has the same cause as the one below.
    """
    ret = salt_ssh_cli.run("pillar.items")
    assert ret.returncode == 0
    assert isinstance(ret.data, dict)
    assert ret.data
    _assert_gpg_pillar(ret.data)


def _assert_gpg_pillar(ret):
    assert "secrets" in ret
    assert "foo" in ret["secrets"]
    assert "BEGIN PGP MESSAGE" not in ret["secrets"]["foo"]
    assert ret["secrets"]["foo"] == "supersecret"
    assert "_errors" not in ret


@pytest.mark.usefixtures("pillar_setup")
def test_saltutil_runner(salt_ssh_cli, salt_minion):
    """
    Ensure that during pillar compilation, the cache dir is not
    overridden. For a history, see PR #50489 and issue #36796,
    notice that the initial description is probably unrelated
    to this.
    """
    ret = salt_ssh_cli.run("pillar.items")
    assert ret.returncode == 0
    assert isinstance(ret.data, dict)
    assert ret.data
    _assert_saltutil_runner_pillar(ret.data, salt_minion.id)


def _assert_saltutil_runner_pillar(ret, salt_minion_id):
    assert "saltutil" in ret
    assert isinstance(ret["saltutil"], dict)
    assert ret["saltutil"]
    assert salt_minion_id in ret["saltutil"]
    assert ret["saltutil"][salt_minion_id] is True
    assert "_errors" not in ret


@pytest.mark.skip_if_binaries_missing("gpg")
@pytest.mark.usefixtures("pillar_setup", "gpg_homedir")
def test_gpg_pillar_orch(salt_ssh_cli, salt_run_cli):
    """
    Ensure that GPG-encrypted pillars can be decrypted when Salt-SSH is
    called during an orchestration or via saltutil.cmd.
    This is issue #65670.
    """
    # Use salt_run_cli since the config paths are different between
    # test master and test minion.
    ret = salt_run_cli.run(
        "salt.cmd",
        "saltutil.cmd",
        salt_ssh_cli.target_host,
        "pillar.items",
        ssh=True,
        roster_file=str(salt_ssh_cli.roster_file),
        ssh_priv=str(salt_ssh_cli.client_key),
    )
    assert ret.returncode == 0
    assert isinstance(ret.data, dict)
    assert ret.data
    _assert_gpg_pillar(ret.data[salt_ssh_cli.target_host]["return"])


@pytest.mark.usefixtures("pillar_setup")
def test_saltutil_runner_orch(salt_ssh_cli, salt_run_cli, salt_minion):
    """
    Ensure that runner calls in the pillar succeed when Salt-SSH is
    called during an orchestration or via saltutil.cmd.
    This is a variant of issue #65670.
    """
    # Use salt_run_cli since the config paths are different between
    # test master and test minion.
    ret = salt_run_cli.run(
        "salt.cmd",
        "saltutil.cmd",
        salt_ssh_cli.target_host,
        "pillar.items",
        ssh=True,
        roster_file=str(salt_ssh_cli.roster_file),
        ssh_priv=str(salt_ssh_cli.client_key),
    )
    assert ret.returncode == 0
    assert isinstance(ret.data, dict)
    assert ret.data
    _assert_saltutil_runner_pillar(
        ret.data[salt_ssh_cli.target_host]["return"], salt_minion.id
    )



"""
Manage Windows Packages using Chocolatey
.. versionadded:: 2016.3.0

.. note::
    Chocolatey pulls data from the Chocolatey internet database to determine
    current versions, find available versions, etc. This is normally a slow
    operation and may be optimized by specifying a local, smaller chocolatey
    repo.
"""

import salt.utils.data
import salt.utils.versions
from salt.exceptions import SaltInvocationError


def __virtual__():
    """
    Load only if chocolatey is loaded
    """
    if "chocolatey.install" in __salt__:
        return "chocolatey"
    return False, "chocolatey module could not be loaded"


def installed(
    name,
    version=None,
    source=None,
    force=False,
    pre_versions=False,
    install_args=None,
    override_args=False,
    force_x86=False,
    package_args=None,
    allow_multiple=False,
    execution_timeout=None,
):
    """
    Installs a package if not already installed

    Args:

        name (str):
            The name of the package to be installed. Required.

        version (str):
            Install a specific version of the package. Defaults to the latest
            version. If the version is different to the one installed, then the
            specified version will be installed. Default is ``None``.

        source (str):
            Chocolatey repository (directory, share or remote URL, feed).
            ``None`` defaults to the official Chocolatey feed. Default is
            ``None``.

        force (bool):
            Reinstall the current version of an existing package. Do not use
            with ``allow_multiple``. Default is ``False``.

        pre_versions (bool):
            Include pre-release packages. Default is ``False``.

        install_args (str):
            Install arguments you want to pass to the installation process, i.e.
            product key or feature list. Default is ``None``.

        override_args (bool):
            Set to ``True`` to override the original install arguments (for the
            native installer) in the package and use your own. When this is set
            to ``False``, install_args will be appended to the end of the
            default arguments. Default is ``False``.

        force_x86 (bool):
            Force x86 (32bit) installation on 64bit systems. Default is
            ``False``.

        package_args (str):
            Arguments you want to pass to the package. Default is ``None``.

        allow_multiple (bool):
            Allow multiple versions of the package to be installed. Do not use
            with ``force``. Does not work with all packages. Default is
            ``False``.

            .. versionadded:: 2017.7.0

        execution_timeout (str):
            Chocolatey execution timeout value you want to pass to the
            installation process. Default is ``None``.

    Example:

    .. code-block:: yaml

        install_some_package:
          chocolatey.installed:
            - name: packagename
            - version: '12.04'
            - source: 'mychocolatey/source'
            - force: True
    """
    if force and allow_multiple:
        raise SaltInvocationError(
            "Cannot use 'force' in conjunction with 'allow_multiple'"
        )

    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    # Get list of currently installed packages
    pre_install = __salt__["chocolatey.list"](local_only=True)

    # Determine action
    # Package not installed
    if name.lower() not in [package.lower() for package in pre_install.keys()]:
        if version:
            ret["comment"] = f"{name} {version} will be installed"
        else:
            ret["comment"] = f"Latest version of {name} will be installed"

    # Package installed
    else:
        full_name = name
        for pkg in pre_install.keys():
            if name.lower() == pkg.lower():
                full_name = pkg

        installed_version = pre_install[full_name]

        if version:
            if salt.utils.versions.compare(
                ver1=installed_version, oper="==", ver2=version
            ):
                if force:
                    ret["comment"] = f"{name} {version} will be reinstalled"
                else:
                    ret["comment"] = f"{name} {version} is already installed"
            else:
                if allow_multiple:
                    ret["comment"] = (
                        f"{name} {version} will be installed side by side with {name} {installed_version} if supported"
                    )
                else:
                    ret["comment"] = (
                        f"{name} {version} will be installed over {name} {installed_version}"
                    )
                    force = True
        else:
            version = installed_version
            if force:
                ret["comment"] = f"{name} {version} will be reinstalled"
            else:
                ret["comment"] = f"{name} {version} is already installed"
                return ret

    if __opts__["test"]:
        ret["result"] = None
        return ret

    # Install the package
    result = __salt__["chocolatey.install"](
        name=name,
        version=version,
        source=source,
        force=force,
        pre_versions=pre_versions,
        install_args=install_args,
        override_args=override_args,
        force_x86=force_x86,
        package_args=package_args,
        allow_multiple=allow_multiple,
        execution_timeout=execution_timeout,
    )

    if "Running chocolatey failed" not in result:
        ret["comment"] = f"{name} installed successfully"
        ret["result"] = True
    else:
        ret["comment"] = f"Failed to install {name}"
        ret["result"] = False

    # Get list of installed packages after 'chocolatey.install'
    post_install = __salt__["chocolatey.list"](local_only=True)

    ret["changes"] = salt.utils.data.compare_dicts(pre_install, post_install)

    return ret


def uninstalled(name, version=None, uninstall_args=None, override_args=False):
    """
    Uninstalls a chocolatey package

    Args:

        name (str):
            The name of the package to be uninstalled. Required.

        version (str):
            Uninstalls a specific version of the package. Defaults to the latest
            version installed.

        uninstall_args (str):
            A list of uninstall arguments you want to pass to the uninstallation
            process, i.e. product key or feature list

        override_args (str):
            Set to ``True`` if you want to override the original uninstall
            arguments (for the native uninstaller) in the package and use your
            own. When this is set to ``False``, uninstall_args will be appended
            to the end of the default arguments

    Example:

    .. code-block:: yaml

      remove_my_package:
        chocolatey.uninstalled:
          - name: mypackage
          - version: '21.5'
    """

    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    # Get list of currently installed packages
    pre_uninstall = __salt__["chocolatey.list"](local_only=True)

    # Determine if package is installed
    if name.lower() in [package.lower() for package in pre_uninstall.keys()]:
        try:
            ret["comment"] = f"{name} {pre_uninstall[name][0]} will be removed"
        except KeyError:
            ret["comment"] = f"{name} will be removed"
    else:
        ret["comment"] = f"The package {name} is not installed"
        return ret

    if __opts__["test"]:
        ret["result"] = None
        return ret

    # Uninstall the package
    result = __salt__["chocolatey.uninstall"](
        name, version, uninstall_args, override_args
    )

    if "Running chocolatey failed" not in result:
        ret["comment"] = f"{name} uninstalled successfully"
        ret["result"] = True
    else:
        ret["comment"] = f"Failed to uninstall {name}"
        ret["result"] = False

    # Get list of installed packages after 'chocolatey.uninstall'
    post_uninstall = __salt__["chocolatey.list"](local_only=True)

    ret["changes"] = salt.utils.data.compare_dicts(pre_uninstall, post_uninstall)

    return ret


def upgraded(
    name,
    version=None,
    source=None,
    force=False,
    pre_versions=False,
    install_args=None,
    override_args=False,
    force_x86=False,
    package_args=None,
):
    """
    Upgrades a chocolatey package. Will install the package if not installed.

    .. versionadded:: 2018.3.0

    Args:

        name (str):
            The name of the package to be installed. Required.

        version (str):
            Install a specific version of the package. Defaults to latest
            version. If the version is greater than the one installed then the
            specified version will be installed. Default is ``None``.

        source (str):
            Chocolatey repository (directory, share or remote URL, feed).
            Defaults to the official Chocolatey feed. Default is ``None``.

        force (bool):
            ``True`` will reinstall an existing package with the same version.
            Default is ``False``.

        pre_versions (bool):
            ``True`` will include pre-release packages. Default is ``False``.

        install_args (str):
            Install arguments you want to pass to the installation process, i.e
            product key or feature list. Default is ``None``.

        override_args (bool):
            ``True`` will override the original install arguments (for the
            native installer) in the package and use those specified in
            ``install_args``. ``False`` will append install_args to the end of
            the default arguments. Default is ``False``.

        force_x86 (bool):
            ``True`` forces 32bit installation on 64bit systems. Default is
            ``False``.

        package_args (str):
            Arguments you want to pass to the package. Default is ``None``.

    Example:

    .. code-block:: yaml

        upgrade_some_package:
          chocolatey.upgraded:
            - name: packagename
            - version: '12.04'
            - source: 'mychocolatey/source'
    """
    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    # Get list of currently installed packages
    pre_install = __salt__["chocolatey.list"](local_only=True)

    # Determine if there are changes
    # Package not installed
    if name.lower() not in [package.lower() for package in pre_install.keys()]:
        if version:
            ret["comment"] = f"{name} {version} will be installed"
        else:
            ret["comment"] = f"Latest version of {name} will be installed"

    # Package installed
    else:
        version_info = __salt__["chocolatey.version"](
            name=name, check_remote=True, source=source
        )

        # Get the actual full name out of version_info
        full_name = name
        for pkg in version_info:
            if name.lower() == pkg.lower():
                full_name = pkg

        installed_version = version_info[full_name]["installed"][0]

        # If version is not passed, use available... if available is available
        if not version:
            if "available" in version_info[full_name]:
                version = version_info[full_name]["available"][0]

        if version:
            # If installed version and new version are the same
            if salt.utils.versions.compare(
                ver1=installed_version, oper="==", ver2=version
            ):
                if force:
                    ret["comment"] = f"{name} {version} will be reinstalled"
                else:
                    ret["comment"] = f"{name} {version} is already installed"
                    return ret
            else:
                # If installed version is older than new version
                if salt.utils.versions.compare(
                    ver1=installed_version, oper="<", ver2=version
                ):
                    ret["comment"] = (
                        f"{name} {installed_version} will be upgraded to version {version}"
                    )
                # If installed version is newer than new version
                else:
                    ret["comment"] = (
                        f"{name} {installed_version} (newer) is already installed"
                    )
                    return ret
        # Catch all for a condition where version is not passed and there is no
        # available version
        else:
            ret["comment"] = "No version found to install"
            return ret

    # Return if running in test mode
    if __opts__["test"]:
        ret["result"] = None
        return ret

    # Install the package
    result = __salt__["chocolatey.upgrade"](
        name=name,
        version=version,
        source=source,
        force=force,
        pre_versions=pre_versions,
        install_args=install_args,
        override_args=override_args,
        force_x86=force_x86,
        package_args=package_args,
    )

    if "Running chocolatey failed" not in result:
        ret["comment"] = f"{name} upgraded successfully"
        ret["result"] = True
    else:
        ret["comment"] = f"Failed to upgrade {name}"
        ret["result"] = False

    # Get list of installed packages after 'chocolatey.install'
    post_install = __salt__["chocolatey.list"](local_only=True)

    # Prior to this, ret['changes'] would have contained expected changes,
    # replace them with the actual changes now that we have completed the
    # installation.
    ret["changes"] = salt.utils.data.compare_dicts(pre_install, post_install)

    return ret


def source_present(
    name, source_location, username=None, password=None, force=False, priority=None
):
    """
    Adds a Chocolatey source if not already present.

    Args:

        name (str):
            The name of the source to be added as a chocolatey repository.

        source (str):
            Location of the source you want to work with.

        username (str):
            The username for a chocolatey source that needs authentication
            credentials.

        password (str):
            The password for a chocolatey source that needx authentication
            credentials.

        force (bool):
            Salt will not modify an existing repository with the same name. Set
            this option to ``True`` to update an existing repository.

        priority (int):
            The priority order of this source as compared to other sources.
            Lower is better. Defaults to 0 (no priority). All priorities
            above 0 will be evaluated first, then zero-based values will be
            evaluated in config file order.

    Example:

    .. code-block:: yaml

        add_some_source:
          chocolatey.source_present:
            - name: reponame
            - source: https://repo.exemple.com
            - username: myuser
            - password: mypassword
            - priority: 100
    """
    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    # Get list of currently present sources
    pre_install = __salt__["chocolatey.list_sources"]()

    # Determine action
    # Source with same name not present
    if name.lower() not in [present.lower() for present in pre_install.keys()]:
        ret["comment"] = f"{name} will be added"

    # Source with same name already present
    else:
        if force:
            ret["comment"] = f"{name} will be updated"
        else:
            ret["comment"] = f"{name} is already present"
            return ret

    if __opts__["test"]:
        ret["result"] = None
        return ret

    # Add the source
    result = __salt__["chocolatey.add_source"](
        name=name,
        source_location=source_location,
        username=username,
        password=password,
        priority=priority,
    )

    if "Running chocolatey failed" not in result:
        ret["result"] = True
        ret["comment"] = f"Source {name} added successfully"
    else:
        ret["result"] = False
        ret["comment"] = f"Failed to add the source {name}"

    # Get list of present sources after 'chocolatey.add_source'
    post_install = __salt__["chocolatey.list_sources"]()

    ret["changes"] = salt.utils.data.compare_dicts(pre_install, post_install)

    return ret



"""
Tests for the salt-run command
"""

import pytest

from tests.support.case import ShellCase


@pytest.mark.windows_whitelisted
@pytest.mark.usefixtures("salt_sub_minion")
class ManageTest(ShellCase):
    """
    Test the manage runner
    """

    @pytest.mark.slow_test
    def test_up(self):
        """
        manage.up
        """
        ret = self.run_run_plus("manage.up", timeout=60)
        self.assertIn("minion", ret["return"])
        self.assertIn("sub_minion", ret["return"])
        self.assertTrue(any("- minion" in out for out in ret["out"]))
        self.assertTrue(any("- sub_minion" in out for out in ret["out"]))

    @pytest.mark.slow_test
    def test_down(self):
        """
        manage.down
        """
        ret = self.run_run_plus("manage.down", timeout=60)
        self.assertNotIn("minion", ret["return"])
        self.assertNotIn("sub_minion", ret["return"])
        self.assertNotIn("minion", ret["out"])
        self.assertNotIn("sub_minion", ret["out"])



"""
Aptly Debian repository manager.

.. versionadded:: 2018.3.0
"""

import logging
import os
import re

import salt.utils.json
import salt.utils.path
import salt.utils.stringutils
from salt.exceptions import SaltInvocationError

_DEFAULT_CONFIG_PATH = "/etc/aptly.conf"
log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "aptly"


def __virtual__():
    """
    Only works on systems with the aptly binary in the system path.
    """
    if salt.utils.path.which("aptly"):
        return __virtualname__
    return (False, "The aptly binaries required cannot be found or are not installed.")


def _cmd_run(cmd):
    """
    Run the aptly command.

    :return: The string output of the command.
    :rtype: str
    """
    cmd.insert(0, "aptly")
    cmd_ret = __salt__["cmd.run_all"](cmd, ignore_retcode=True)

    if cmd_ret["retcode"] != 0:
        log.debug("Unable to execute command: %s\nError: %s", cmd, cmd_ret["stderr"])

    return cmd_ret["stdout"]


def _format_repo_args(
    comment=None, component=None, distribution=None, uploaders_file=None, saltenv="base"
):
    """
    Format the common arguments for creating or editing a repository.

    :param str comment: The description of the repository.
    :param str component: The default component to use when publishing.
    :param str distribution: The default distribution to use when publishing.
    :param str uploaders_file: The repository upload restrictions config.
    :param str saltenv: The environment the file resides in.

    :return: A list of the arguments formatted as aptly arguments.
    :rtype: list
    """
    ret = list()
    cached_uploaders_path = None
    settings = {
        "comment": comment,
        "component": component,
        "distribution": distribution,
    }

    if uploaders_file:
        cached_uploaders_path = __salt__["cp.cache_file"](uploaders_file, saltenv)

        if not cached_uploaders_path:
            log.error("Unable to get cached copy of file: %s", uploaders_file)
            return False

    for setting in settings:
        if settings[setting] is not None:
            ret.append(f"-{setting}={settings[setting]}")

    if cached_uploaders_path:
        ret.append(f"-uploaders-file={cached_uploaders_path}")

    return ret


def _validate_config(config_path):
    """
    Validate that the configuration file exists and is readable.

    :param str config_path: The path to the configuration file for the aptly instance.

    :return: None
    :rtype: None
    """
    log.debug("Checking configuration file: %s", config_path)

    if not os.path.isfile(config_path):
        message = f"Unable to get configuration file: {config_path}"
        log.error(message)
        raise SaltInvocationError(message)


def get_config(config_path=_DEFAULT_CONFIG_PATH):
    """
    Get the configuration data.

    :param str config_path: The path to the configuration file for the aptly instance.

    :return: A dictionary containing the configuration data.
    :rtype: dict

    CLI Example:

    .. code-block:: bash

        salt '*' aptly.get_config
    """
    _validate_config(config_path)

    cmd = ["config", "show", f"-config={config_path}"]

    cmd_ret = _cmd_run(cmd)

    return salt.utils.json.loads(cmd_ret)


def list_repos(config_path=_DEFAULT_CONFIG_PATH, with_packages=False):
    """
    List all of the repos.

    :param str config_path: The path to the configuration file for the aptly instance.
    :param bool with_packages: Return a list of packages in the repo.

    :return: A dictionary of the repositories.
    :rtype: dict

    CLI Example:

    .. code-block:: bash

        salt '*' aptly.list_repos
    """
    _validate_config(config_path)

    ret = dict()
    cmd = ["repo", "list", f"-config={config_path}", "-raw=true"]

    cmd_ret = _cmd_run(cmd)
    repos = [line.strip() for line in cmd_ret.splitlines()]

    log.debug("Found repositories: %s", len(repos))

    for name in repos:
        ret[name] = get_repo(
            name=name, config_path=config_path, with_packages=with_packages
        )
    return ret


def get_repo(name, config_path=_DEFAULT_CONFIG_PATH, with_packages=False):
    """
    Get the details of the repository.

    :param str name: The name of the repository.
    :param str config_path: The path to the configuration file for the aptly instance.
    :param bool with_packages: Return a list of packages in the repo.

    :return: A dictionary containing information about the repository.
    :rtype: dict

    CLI Example:

    .. code-block:: bash

        salt '*' aptly.get_repo name="test-repo"
    """
    _validate_config(config_path)
    with_packages = str(bool(with_packages)).lower()

    ret = dict()
    cmd = [
        "repo",
        "show",
        f"-config={config_path}",
        f"-with-packages={with_packages}",
        name,
    ]

    cmd_ret = _cmd_run(cmd)

    for line in cmd_ret.splitlines():
        try:
            # Extract the settings and their values, and attempt to format
            # them to match their equivalent setting names.
            items = line.split(":")
            key = items[0].lower().replace("default", "").strip()
            key = " ".join(key.split()).replace(" ", "_")
            ret[key] = salt.utils.stringutils.to_none(
                salt.utils.stringutils.to_num(items[1].strip())
            )
        except (AttributeError, IndexError):
            # If the line doesn't have the separator or is otherwise invalid, skip it.
            log.debug("Skipping line: %s", line)

    if ret:
        log.debug("Found repository: %s", name)
    else:
        log.debug("Unable to find repository: %s", name)
    return ret


def new_repo(
    name,
    config_path=_DEFAULT_CONFIG_PATH,
    comment=None,
    component=None,
    distribution=None,
    uploaders_file=None,
    from_snapshot=None,
    saltenv="base",
):
    """
    Create the new repository.

    :param str name: The name of the repository.
    :param str config_path: The path to the configuration file for the aptly instance.
    :param str comment: The description of the repository.
    :param str component: The default component to use when publishing.
    :param str distribution: The default distribution to use when publishing.
    :param str uploaders_file: The repository upload restrictions config.
    :param str from_snapshot: The snapshot to initialize the repository contents from.
    :param str saltenv: The environment the file resides in.

    :return: A boolean representing whether all changes succeeded.
    :rtype: bool

    CLI Example:

    .. code-block:: bash

        salt '*' aptly.new_repo name="test-repo" comment="Test main repo" component="main" distribution="trusty"
    """
    _validate_config(config_path)

    current_repo = __salt__["aptly.get_repo"](name=name, config_path=config_path)

    if current_repo:
        log.debug("Repository already exists: %s", name)
        return True

    cmd = ["repo", "create", f"-config={config_path}"]
    repo_params = _format_repo_args(
        comment=comment,
        component=component,
        distribution=distribution,
        uploaders_file=uploaders_file,
        saltenv=saltenv,
    )
    cmd.extend(repo_params)
    cmd.append(name)

    if from_snapshot:
        cmd.extend(["from", "snapshot", from_snapshot])

    _cmd_run(cmd)
    repo = __salt__["aptly.get_repo"](name=name, config_path=config_path)

    if repo:
        log.debug("Created repo: %s", name)
        return True
    log.error("Unable to create repo: %s", name)
    return False


def set_repo(
    name,
    config_path=_DEFAULT_CONFIG_PATH,
    comment=None,
    component=None,
    distribution=None,
    uploaders_file=None,
    saltenv="base",
):
    """
    Configure the repository settings.

    :param str name: The name of the repository.
    :param str config_path: The path to the configuration file for the aptly instance.
    :param str comment: The description of the repository.
    :param str component: The default component to use when publishing.
    :param str distribution: The default distribution to use when publishing.
    :param str uploaders_file: The repository upload restrictions config.
    :param str from_snapshot: The snapshot to initialize the repository contents from.
    :param str saltenv: The environment the file resides in.

    :return: A boolean representing whether all changes succeeded.
    :rtype: bool

    CLI Example:

    .. code-block:: bash

        salt '*' aptly.set_repo name="test-repo" comment="Test universe repo" component="universe" distribution="xenial"
    """
    _validate_config(config_path)

    failed_settings = dict()

    # Only check for settings that were passed in and skip the rest.
    settings = {
        "comment": comment,
        "component": component,
        "distribution": distribution,
    }

    for setting in list(settings):
        if settings[setting] is None:
            settings.pop(setting, None)

    current_settings = __salt__["aptly.get_repo"](name=name, config_path=config_path)

    if not current_settings:
        log.error("Unable to get repo: %s", name)
        return False

    # Discard any additional settings that get_repo gives
    # us that are not present in the provided arguments.
    for current_setting in list(current_settings):
        if current_setting not in settings:
            current_settings.pop(current_setting, None)

    # Check the existing repo settings to see if they already have the desired values.
    if settings == current_settings:
        log.debug("Settings already have the desired values for repository: %s", name)
        return True

    cmd = ["repo", "edit", f"-config={config_path}"]

    repo_params = _format_repo_args(
        comment=comment,
        component=component,
        distribution=distribution,
        uploaders_file=uploaders_file,
        saltenv=saltenv,
    )
    cmd.extend(repo_params)
    cmd.append(name)

    _cmd_run(cmd)
    new_settings = __salt__["aptly.get_repo"](name=name, config_path=config_path)

    # Check the new repo settings to see if they have the desired values.
    for setting in settings:
        if settings[setting] != new_settings[setting]:
            failed_settings.update({setting: settings[setting]})

    if failed_settings:
        log.error("Unable to change settings for the repository: %s", name)
        return False
    log.debug(
        "Settings successfully changed to the desired values for repository: %s", name
    )
    return True


def delete_repo(name, config_path=_DEFAULT_CONFIG_PATH, force=False):
    """
    Remove the repository.

    :param str name: The name of the repository.
    :param str config_path: The path to the configuration file for the aptly instance.
    :param bool force: Whether to remove the repository even if it is used as the source
        of an existing snapshot.

    :return: A boolean representing whether all changes succeeded.
    :rtype: bool

    CLI Example:

    .. code-block:: bash

        salt '*' aptly.delete_repo name="test-repo"
    """
    _validate_config(config_path)
    force = str(bool(force)).lower()

    current_repo = __salt__["aptly.get_repo"](name=name, config_path=config_path)

    if not current_repo:
        log.debug("Repository already absent: %s", name)
        return True

    cmd = [
        "repo",
        "drop",
        f"-config={config_path}",
        f"-force={force}",
        name,
    ]

    _cmd_run(cmd)
    repo = __salt__["aptly.get_repo"](name=name, config_path=config_path)

    if repo:
        log.error("Unable to remove repo: %s", name)
        return False
    log.debug("Removed repo: %s", name)
    return True


def list_mirrors(config_path=_DEFAULT_CONFIG_PATH):
    """
    Get a list of all the mirrors.

    :param str config_path: The path to the configuration file for the aptly instance.

    :return: A list of the mirror names.
    :rtype: list

    CLI Example:

    .. code-block:: bash

        salt '*' aptly.list_mirrors
    """
    _validate_config(config_path)

    cmd = ["mirror", "list", f"-config={config_path}", "-raw=true"]

    cmd_ret = _cmd_run(cmd)
    ret = [line.strip() for line in cmd_ret.splitlines()]

    log.debug("Found mirrors: %s", len(ret))
    return ret


def list_published(config_path=_DEFAULT_CONFIG_PATH):
    """
    Get a list of all the published repositories.

    :param str config_path: The path to the configuration file for the aptly instance.

    :return: A list of the published repository names.
    :rtype: list

    CLI Example:

    .. code-block:: bash

        salt '*' aptly.list_published
    """
    _validate_config(config_path)

    cmd = ["publish", "list", f"-config={config_path}", "-raw=true"]

    cmd_ret = _cmd_run(cmd)
    ret = [line.strip() for line in cmd_ret.splitlines()]

    log.debug("Found published repositories: %s", len(ret))
    return ret


def list_snapshots(config_path=_DEFAULT_CONFIG_PATH, sort_by_time=False):
    """
    Get a list of all the snapshots.

    :param str config_path: The path to the configuration file for the aptly instance.
    :param bool sort_by_time: Whether to sort by creation time instead of by name.

    :return: A list of the snapshot names.
    :rtype: list

    CLI Example:

    .. code-block:: bash

        salt '*' aptly.list_snapshots
    """
    _validate_config(config_path)

    cmd = ["snapshot", "list", f"-config={config_path}", "-raw=true"]

    if sort_by_time:
        cmd.append("-sort=time")
    else:
        cmd.append("-sort=name")

    cmd_ret = _cmd_run(cmd)
    ret = [line.strip() for line in cmd_ret.splitlines()]

    log.debug("Found snapshots: %s", len(ret))
    return ret


def cleanup_db(config_path=_DEFAULT_CONFIG_PATH, dry_run=False):
    """
    Remove data regarding unreferenced packages and delete files in the package pool that
        are no longer being used by packages.

    :param bool dry_run: Report potential changes without making any changes.

    :return: A dictionary of the package keys and files that were removed.
    :rtype: dict

    CLI Example:

    .. code-block:: bash

        salt '*' aptly.cleanup_db
    """
    _validate_config(config_path)
    dry_run = str(bool(dry_run)).lower()

    ret = {"deleted_keys": list(), "deleted_files": list()}

    cmd = [
        "db",
        "cleanup",
        f"-config={config_path}",
        f"-dry-run={dry_run}",
        "-verbose=true",
    ]

    cmd_ret = _cmd_run(cmd)

    type_pattern = r"^List\s+[\w\s]+(?P<package_type>(file|key)s)[\w\s]+:$"
    list_pattern = r"^\s+-\s+(?P<package>.*)$"
    current_block = None

    for line in cmd_ret.splitlines():
        if current_block:
            match = re.search(list_pattern, line)
            if match:
                package_type = f"deleted_{current_block}"
                ret[package_type].append(match.group("package"))
            else:
                current_block = None
        # Intentionally not using an else here, in case of a situation where
        # the next list header might be bordered by the previous list.
        if not current_block:
            match = re.search(type_pattern, line)
            if match:
                current_block = match.group("package_type")

    log.debug("Package keys identified for deletion: %s", len(ret["deleted_keys"]))
    log.debug("Package files identified for deletion: %s", len(ret["deleted_files"]))
    return ret



"""
Test case for the vault execution module
"""

import pytest

import salt.modules.vault as vault
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {
        vault: {
            "__grains__": {"id": "foo"},
            "__utils__": {
                "vault.is_v2": MagicMock(
                    return_value={
                        "v2": True,
                        "data": "secrets/data/mysecret",
                        "metadata": "secrets/metadata/mysecret",
                        "type": "kv",
                    }
                ),
            },
        },
    }


@pytest.fixture
def path():
    return "foo/bar/"


def test_read_secret_v1():
    """
    Test salt.modules.vault.read_secret function
    """
    version = {"v2": False, "data": None, "metadata": None, "type": None}
    mock_version = MagicMock(return_value=version)
    mock_vault = MagicMock()
    mock_vault.return_value.status_code = 200
    mock_vault.return_value.json.return_value = {"data": {"key": "test"}}
    with patch.dict(vault.__utils__, {"vault.make_request": mock_vault}), patch.dict(
        vault.__utils__, {"vault.is_v2": mock_version}
    ):
        vault_return = vault.read_secret("/secret/my/secret")

    assert vault_return == {"key": "test"}


def test_read_secret_v1_key():
    """
    Test salt.modules.vault.read_secret function specifying key
    """
    version = {"v2": False, "data": None, "metadata": None, "type": None}
    mock_version = MagicMock(return_value=version)
    mock_vault = MagicMock()
    mock_vault.return_value.status_code = 200
    mock_vault.return_value.json.return_value = {"data": {"key": "somevalue"}}
    with patch.dict(vault.__utils__, {"vault.make_request": mock_vault}), patch.dict(
        vault.__utils__, {"vault.is_v2": mock_version}
    ):
        vault_return = vault.read_secret("/secret/my/secret", "key")

    assert vault_return == "somevalue"


def test_read_secret_v2():
    """
    Test salt.modules.vault.read_secret function for v2 of kv secret backend
    """
    # given path secrets/mysecret generate v2 output
    version = {
        "v2": True,
        "data": "secrets/data/mysecret",
        "metadata": "secrets/metadata/mysecret",
        "type": "kv",
    }
    mock_version = MagicMock(return_value=version)
    mock_vault = MagicMock()
    mock_vault.return_value.status_code = 200
    v2_return = {
        "data": {
            "data": {"akey": "avalue"},
            "metadata": {
                "created_time": "2018-10-23T20:21:55.042755098Z",
                "destroyed": False,
                "version": 13,
                "deletion_time": "",
            },
        }
    }

    mock_vault.return_value.json.return_value = v2_return
    with patch.dict(vault.__utils__, {"vault.make_request": mock_vault}), patch.dict(
        vault.__utils__, {"vault.is_v2": mock_version}
    ):
        # Validate metadata returned
        vault_return = vault.read_secret("/secret/my/secret", metadata=True)
        assert "data" in vault_return
        assert "metadata" in vault_return
        # Validate just data returned
        vault_return = vault.read_secret("/secret/my/secret")
        assert "akey" in vault_return


def test_read_secret_v2_key():
    """
    Test salt.modules.vault.read_secret function for v2 of kv secret backend
    with specified key
    """
    # given path secrets/mysecret generate v2 output
    version = {
        "v2": True,
        "data": "secrets/data/mysecret",
        "metadata": "secrets/metadata/mysecret",
        "type": "kv",
    }
    mock_version = MagicMock(return_value=version)
    mock_vault = MagicMock()
    mock_vault.return_value.status_code = 200
    v2_return = {
        "data": {
            "data": {"akey": "avalue"},
            "metadata": {
                "created_time": "2018-10-23T20:21:55.042755098Z",
                "destroyed": False,
                "version": 13,
                "deletion_time": "",
            },
        }
    }

    mock_vault.return_value.json.return_value = v2_return
    with patch.dict(vault.__utils__, {"vault.make_request": mock_vault}), patch.dict(
        vault.__utils__, {"vault.is_v2": mock_version}
    ):
        vault_return = vault.read_secret("/secret/my/secret", "akey")

    assert vault_return == "avalue"


def test_read_secret_with_default(path):
    assert vault.read_secret(path, default="baz") == "baz"


def test_read_secret_no_default(path):
    with pytest.raises(CommandExecutionError):
        vault.read_secret(path)


def test_list_secrets_with_default(path):
    assert vault.list_secrets(path, default=["baz"]) == ["baz"]


def test_list_secrets_no_default(path):
    with pytest.raises(CommandExecutionError):
        vault.list_secrets(path)



import os

import pytest

pytestmark = [
    pytest.mark.windows_whitelisted,
]


def test_replace(file, tmp_path):
    """
    file.replace
    """
    name = tmp_path / "testfile"
    name.write_text("change_me")

    ret = file.replace(name=str(name), pattern="change", repl="salt", backup=False)
    assert ret.result is True
    assert name.read_text() == "salt_me"


def test_replace_issue_18612(file, tmp_path):
    """
    Test the (mis-)behaviour of file.replace as described in #18612:

    Using 'prepend_if_not_found' or 'append_if_not_found' resulted in
    an infinitely growing file as 'file.replace' didn't check beforehand
    whether the changes had already been done to the file

    # Case description:

    The tested file contains one commented line
    The commented line should be uncommented in the end, nothing else should change
    """
    test_name = "test_replace_issue_18612"
    path_test = tmp_path / test_name
    path_test.write_text("# en_US.UTF-8")

    ret = []
    for _ in range(3):
        ret.append(
            file.replace(
                name=str(path_test),
                pattern="^# en_US.UTF-8$",
                repl="en_US.UTF-8",
                append_if_not_found=True,
            )
        )

    # ensure, the number of lines didn't change, even after invoking 'file.replace' 3 times
    # and that the replacement succeeded
    assert path_test.read_text() == "en_US.UTF-8"

    # ensure, all runs of 'file.replace' reported success
    for item in ret:
        assert item.result is True


def test_replace_issue_18612_prepend(file, tmp_path):
    """
    Test the (mis-)behaviour of file.replace as described in #18612:

    Using 'prepend_if_not_found' or 'append_if_not_found' resulted in
    an infinitely growing file as 'file.replace' didn't check beforehand
    whether the changes had already been done to the file

    # Case description:

    The tested file contains multiple lines not matching the pattern or replacement in any way
    The replacement pattern should be prepended to the file
    """
    test_name = "test_replace_issue_18612_prepend"
    path_test = tmp_path / test_name
    contents = "#john=doe\n#foo = bar\n#foo=bar\n#blah=blub\n#some=thing"
    path_test.write_text(contents)

    ret = []
    for _ in range(3):
        ret.append(
            file.replace(
                name=str(path_test),
                pattern="^# en_US.UTF-8$",
                repl="en_US.UTF-8",
                prepend_if_not_found=True,
            )
        )

    # ensure, the resulting file contains the expected lines
    assert path_test.read_text() == f"en_US.UTF-8\n{contents}"

    backup_file = path_test.with_name(f"{path_test.name}.bak")
    assert backup_file.is_file()
    assert backup_file.read_text() == contents

    # ensure, all runs of 'file.replace' reported success
    for item in ret:
        assert item.result is True


def test_replace_issue_18612_append(file, tmp_path):
    """
    Test the (mis-)behaviour of file.replace as described in #18612:

    Using 'prepend_if_not_found' or 'append_if_not_found' resulted in
    an infinitely growing file as 'file.replace' didn't check beforehand
    whether the changes had already been done to the file

    # Case description:

    The tested file contains multiple lines not matching the pattern or replacement in any way
    The replacement pattern should be appended to the file
    """
    test_name = "test_replace_issue_18612_append"
    path_test = tmp_path / test_name
    contents = "#john=doe\n#foo = bar\n#foo=bar\n#blah=blub\n#some=thing"
    path_test.write_text(contents)

    ret = []
    for _ in range(3):
        ret.append(
            file.replace(
                name=str(path_test),
                pattern="^# en_US.UTF-8$",
                repl="en_US.UTF-8",
                append_if_not_found=True,
            )
        )

    # ensure, the resulting file contains the expected lines
    assert path_test.read_text() == f"{contents}\nen_US.UTF-8\n"

    backup_file = path_test.with_name(f"{path_test.name}.bak")
    assert backup_file.is_file()
    assert backup_file.read_text() == contents

    # ensure, all runs of 'file.replace' reported success
    for item in ret:
        assert item.result is True


def test_replace_issue_18612_append_not_found_content(file, tmp_path):
    """
    Test the (mis-)behaviour of file.replace as described in #18612:

    Using 'prepend_if_not_found' or 'append_if_not_found' resulted in
    an infinitely growing file as 'file.replace' didn't check beforehand
    whether the changes had already been done to the file

    # Case description:

    The tested file contains multiple lines not matching the pattern or replacement in any way
    The 'not_found_content' value should be appended to the file
    """
    test_name = "test_replace_issue_18612_append_not_found_content"
    path_test = tmp_path / test_name
    contents = "#john=doe\n#foo = bar\n#foo=bar\n#blah=blub\n#some=thing"
    path_test.write_text(contents)

    not_found_content = "THIS LINE WASN'T FOUND! SO WE'RE APPENDING IT HERE!"

    ret = []
    for _ in range(3):
        ret.append(
            file.replace(
                name=str(path_test),
                pattern="^# en_US.UTF-8$",
                repl="en_US.UTF-8",
                append_if_not_found=True,
                not_found_content=not_found_content,
            )
        )

    # ensure, the resulting file contains the expected lines
    assert path_test.read_text() == f"{contents}\n{not_found_content}\n"

    backup_file = path_test.with_name(f"{path_test.name}.bak")
    assert backup_file.is_file()
    assert backup_file.read_text() == contents

    # ensure, all runs of 'file.replace' reported success
    for item in ret:
        assert item.result is True


def test_replace_issue_18612_change_mid_line_with_comment(file, tmp_path):
    """
    Test the (mis-)behaviour of file.replace as described in #18612:

    Using 'prepend_if_not_found' or 'append_if_not_found' resulted in
    an infinitely growing file as 'file.replace' didn't check beforehand
    whether the changes had already been done to the file

    # Case description:

    The tested file contains 5 key=value pairs
    The commented key=value pair #foo=bar should be changed to foo=salt
    The comment char (#) in front of foo=bar should be removed
    """
    test_name = "test_replace_issue_18612_change_mid_line_with_comment"
    path_test = tmp_path / test_name
    contents = "#john=doe\n#foo = bar\n#foo=bar\n#blah=blub\n#some=thing"
    path_test.write_text(contents)

    ret = []
    for _ in range(3):
        ret.append(
            file.replace(
                name=str(path_test),
                pattern="^#foo=bar($|(?=\r\n))",
                repl="foo=salt",
                append_if_not_found=True,
            )
        )

    # ensure, the resulting file contains the expected lines
    assert path_test.read_text() == contents.replace("#foo=bar", "foo=salt")

    backup_file = path_test.with_name(f"{path_test.name}.bak")
    assert backup_file.is_file()
    assert backup_file.read_text() == contents

    # ensure, all runs of 'file.replace' reported success
    for item in ret:
        assert item.result is True


def test_replace_issue_18841_no_changes(file, tmp_path):
    """
    Test the (mis-)behaviour of file.replace as described in #18841:

    Using file.replace in a way which shouldn't modify the file at all
    results in changed mtime of the original file and a backup file being created.

    # Case description

    The tested file contains multiple lines
    The tested file contains a line already matching the replacement (no change needed)
    The tested file's content shouldn't change at all
    The tested file's mtime shouldn't change at all
    No backup file should be created
    """
    test_name = "test_replace_issue_18841_no_changes"
    path_test = tmp_path / test_name
    contents = "#john=doe\n#foo = bar\ngoodbye world\n#foo=bar\n#blah=blub\n#some=thing"
    path_test.write_text(contents)

    # get (m|a)time of file
    fstats_orig = path_test.stat()

    # define how far we predate the file
    age = 5 * 24 * 60 * 60

    # set (m|a)time of file 5 days into the past
    os.utime(str(path_test), (fstats_orig.st_mtime - age, fstats_orig.st_atime - age))

    # Get stat again after we modified it
    fstats_orig = path_test.stat()

    ret = file.replace(
        name=str(path_test),
        pattern="^hello world$",
        repl="goodbye world",
        show_changes=True,
        flags=["IGNORECASE"],
        backup=False,
    )
    assert ret.result is True

    # get (m|a)time of file
    fstats_post = path_test.stat()

    # ensure, the file content didn't change
    assert path_test.read_text() == contents

    # ensure no backup file was created
    backup_file = path_test.with_name(f"{path_test.name}.bak")
    assert backup_file.is_file() is False

    # ensure the file's mtime didn't change
    assert fstats_post.st_mtime == fstats_orig.st_mtime


def test_replace_issue_18841_omit_backup(file, tmp_path):
    """
    Test the (mis-)behaviour of file.replace as described in #18841:

    Using file.replace in a way which shouldn't modify the file at all
    results in changed mtime of the original file and a backup file being created.

    # Case description

    The tested file contains multiple lines
    The tested file contains a line already matching the replacement (no change needed)
    The tested file's content shouldn't change at all
    The tested file's mtime shouldn't change at all
    No backup file should be created, although backup=False isn't explicitly defined
    """
    test_name = "test_replace_issue_18841_omit_backup"
    path_test = tmp_path / test_name
    contents = "#john=doe\n#foo = bar\ngoodbye world\n#foo=bar\n#blah=blub\n#some=thing"
    path_test.write_text(contents)

    # get (m|a)time of file
    fstats_orig = path_test.stat()

    # define how far we predate the file
    age = 5 * 24 * 60 * 60

    # set (m|a)time of file 5 days into the past
    os.utime(str(path_test), (fstats_orig.st_mtime - age, fstats_orig.st_atime - age))

    # Grab stat again after changing (m|a)time
    # fstats_orig = path_test.stat()

    ret = file.replace(
        name=str(path_test),
        pattern="^hello world$",
        repl="goodbye world",
        show_changes=True,
        flags=["IGNORECASE"],
    )
    assert ret.result is True

    # get (m|a)time of file
    # fstats_post = path_test.stat()

    # ensure, the file content didn't change
    assert path_test.read_text() == contents

    # ensure no backup file was created
    backup_file = path_test.with_name(f"{path_test.name}.bak")
    assert backup_file.is_file() is False

    # ensure the file's mtime didn't change
    # assert fstats_post.st_mtime == fstats_orig.st_mtime - age
    # This is commented out because before the test was migrated, while it was
    # passing, checking mtime was wrongly being done:
    #
    #    self.assertTrue(fstats_post.st_mtime, fstats_orig.st_mtime - age)
    #
    # The above is ALWAYS true. Anyway, the backup file was not created, so the
    # test is still valid and valuable


def test_file_replace_prerequired_issues_55775(modules, state_tree, tmp_path):
    """
    Test that __prerequired__ is filter from file.replace
    if __prerequired__ is not filter from file.replace an error will be raised
    """
    managed_file = tmp_path / "issue-55775.txt"
    sls_contents = f"""
    {managed_file}:
      file.managed:
        - contents:
          - foo
    file.replace:
      file.replace:
        - name: {managed_file}
        - pattern: 'foo'
        - repl: 'bar'
        - prereq:
          - test no changes
          - test changes
    test no changes:
      test.succeed_without_changes:
        - name: no changes
    test changes:
      test.succeed_with_changes:
        - name: changes
        - require:
          - test: test no changes
    """
    with pytest.helpers.temp_file("file-replace.sls", sls_contents, state_tree):
        ret = modules.state.sls("file-replace")
        for state_run in ret:
            assert state_run.result is True

    assert managed_file.exists()


def test_file_replace_check_cmd(modules, state_tree):
    """
    Test that check_cmd works for file.replace
    and those states do not run.
    """
    sls_contents = """
    replace_in_file:
      file.replace:
        - name: /tmp/test
        - pattern: hi
        - repl: "replacement text"
        - append_if_not_found: True
        - check_cmd:
          - "djasjahj"
    """
    with pytest.helpers.temp_file(
        "file-replace-check-cmd.sls", sls_contents, state_tree
    ):
        ret = modules.state.sls("file-replace-check-cmd")
        for state_run in ret:
            assert state_run.result is False
            assert state_run.comment == "check_cmd determined the state failed"



import logging
import os
import shutil

import pytest

import salt.config
import salt.loader
import salt.modules.cmdmod as cmdmod
import salt.modules.file as filemod
import salt.utils.data
import salt.utils.files
import salt.utils.platform
import salt.utils.stringutils
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)

pytestmark = pytest.mark.skipif(
    salt.modules.selinux.getenforce() != "Enforcing",
    reason="Skip if selinux not enabled",
)


@pytest.fixture
def configure_loader_modules():
    return {
        filemod: {
            "__salt__": {
                "cmd.run": cmdmod.run,
                "cmd.run_all": cmdmod.run_all,
                "cmd.retcode": cmdmod.retcode,
                "selinux.fcontext_add_policy": MagicMock(
                    return_value={"retcode": 0, "stdout": ""}
                ),
            },
            "__opts__": {"test": False},
        }
    }


@pytest.fixture
def subdir(tmp_path):
    subdir = tmp_path / "file-selinux-test-dir"
    subdir.mkdir()
    yield subdir
    shutil.rmtree(str(subdir))


@pytest.fixture
def tfile1(subdir):
    filename = str(subdir / "tfile1")
    with salt.utils.files.fopen(filename, "w+"):
        pass
    yield filename
    os.remove(filename)


@pytest.fixture
def tfile2(subdir):
    filename = str(subdir / "tfile2")
    with salt.utils.files.fopen(filename, "w+"):
        pass
    yield filename
    os.remove(filename)


@pytest.fixture
def tfile3(subdir):
    filename = str(subdir / "tfile3")
    with salt.utils.files.fopen(filename, "w+"):
        pass
    yield filename
    os.remove(filename)


def test_selinux_getcontext(tfile1):
    """
    Test get selinux context
    Assumes default selinux attributes on temporary files
    """
    result = filemod.get_selinux_context(tfile1)
    assert result == "unconfined_u:object_r:user_tmp_t:s0"


def test_selinux_setcontext(tfile2):
    """
    Test set selinux context
    Assumes default selinux attributes on temporary files
    """
    result = filemod.set_selinux_context(tfile2, user="system_u")
    assert result == "system_u:object_r:user_tmp_t:s0"


def test_selinux_setcontext_persist(tfile2):
    """
    Test set selinux context with persist=True
    Assumes default selinux attributes on temporary files
    """
    result = filemod.set_selinux_context(tfile2, user="system_u", persist=True)
    assert result == "system_u:object_r:user_tmp_t:s0"


def test_selinux_setcontext_persist_change(tfile2):
    """
    Test set selinux context with persist=True
    Assumes default selinux attributes on temporary files
    """
    result = filemod.set_selinux_context(tfile2, user="system_u", persist=True)
    assert result == "system_u:object_r:user_tmp_t:s0"

    result = filemod.set_selinux_context(
        tfile2, user="unconfined_u", type="net_conf_t", persist=True
    )
    assert result == "unconfined_u:object_r:net_conf_t:s0"


def test_file_check_perms(tfile3):
    expected_result = (
        {
            "comment": f"The file {tfile3} is set to be changed",
            "changes": {
                "selinux": {"New": "Type: lost_found_t", "Old": "Type: user_tmp_t"},
                "mode": "0664",
            },
            "name": tfile3,
            "result": True,
        },
        {"cmode": "0664", "luser": "root", "lmode": "0644", "lgroup": "root"},
    )

    # Disable lsattr calls
    with patch("salt.utils.path.which") as m_which:
        m_which.return_value = None
        result = filemod.check_perms(
            tfile3,
            {},
            "root",
            "root",
            664,
            seuser=None,
            serole=None,
            setype="lost_found_t",
            serange=None,
        )
        assert result == expected_result
